from collections import OrderedDict
from typing import List

import pandas as pd

"""
Handles compilation of data while using the taskId as the primary key for the data
"""
class TaskField:
    def __init__(self, name: str = None):
        if name is None:
            self.name = type(self).__name__
        else:
            self.name = name

    def get(self, task_id: int):
        return task_id


class TaskFieldList(List[TaskField]):
    """List of Field types"""

    def get_names(self):
        return [field.name for field in self]

    def to_data(self, task_ids: list[int]) -> pd.DataFrame:
        """
        Given a list of task_ids, calls all the TaskFields to get the data for each task_id
        and returns a dataframe with the data
        """
        task_list = [Task(task_id, self) for task_id in task_ids]
        data = []
        for i, task in enumerate(task_list):
            if i % 100 == 0:
                print("working on", i, "out of", len(task_list))
            task.append_to_data(data)
        return pd.DataFrame(data)

    def append_to_data(self, data: pd.DataFrame):
        """ Creates a new dataframe from the triallist and appends it to the supplied dataframe
        This is used to add new columns (fields) to an existing dataframe"""
        # get first column
        new_data = []
        task_ids = data[data.columns[0]]
        task_list = [Task(task_id, self) for task_id in task_ids]
        for i, task in enumerate(task_list):
            print("working on", i, "out of", len(task_list))
            task.append_to_data(new_data)

        new_df = pd.DataFrame(new_data, columns=self.get_names())
        data = pd.concat([data, new_df], axis=1)
        return data


class Task:
    def __init__(self, task_id: int, fields: TaskFieldList):
        self.task_id = task_id
        self.fields = fields

    def append_to_data(self, data):
        field_values = []
        for field in self.fields:
            try:
                field_value = field.get(self.task_id)
            except Exception as e:
                error = f"Error getting {field.name} for task_id {self.task_id}: {e}"
                print(error)
                field_value = "None"
            field_values.append(field_value)
        names = self.fields.get_names()
        new_row = OrderedDict(zip(names, field_values))
        data.append(new_row)


def get_data_from_tasks(fields: TaskFieldList, task_ids: list[int]) -> pd.DataFrame:
    task_list = [Task(task_id, fields) for task_id in task_ids]
    data = []
    for i, task in enumerate(task_list):
        print("working on", i, "out of", len(task_list))
        task.append_to_data(data)
    return pd.DataFrame(data)



import mmap
import os

import numpy as np

from clat.intan.channels import Channel  # Assuming you have this import


def read_amplifier_data_with_mmap(file_path, amplifier_channels):
    """
    Reads amplifier data from a binary file and maps it to memory using mmap.
    Creates a dictionary mapping each native_channel_name to its vector of spikes.

    Parameters:
    - file_path: str, path to the amplifier.dat file
    - amplifier_channels: list of dicts, amplifier channel information

    Returns:
    - channel_to_data: dict, mapping from native_channel_name to spike data in microvolts
    """
    # Extract the number of channels
    num_channels = len(amplifier_channels)

    # Get file size in bytes
    fileinfo = os.path.getsize(file_path)

    # Calculate the number of samples
    num_samples = fileinfo // (num_channels * 2)  # int16 = 2 bytes

    # Initialize an empty dictionary to store the mapping
    channel_to_data = {}

    # Open the file and map it to memory
    with open(file_path, 'rb') as f:
        mmapped_file = mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ)

        # Create a NumPy array that views the mmapped data
        data = np.frombuffer(mmapped_file, dtype=np.int16).reshape((num_samples, num_channels))

        # Convert the data to microvolts and populate the dictionary
        for i, channel_info in enumerate(amplifier_channels):
            print(f"Reading channel {i}...")
            native_channel_name = channel_info.get("native_channel_name", f"Channel_{i}")
            channel_data = data[:, i] * 0.195  # Conversion to microvolts
            channel_to_data[Channel(native_channel_name)] = channel_data

    return channel_to_data


def read_amplifier_data_with_memmap(file_path, amplifier_channels):
    """
    Reads amplifier data from a binary file and maps it to memory.
    Creates a dictionary mapping each native_channel_name to its vector of spikes.

    Parameters:
    - file_path: str, path to the amplifier.dat file
    - amplifier_channels: list of dicts, amplifier channel information

    Returns:
    - channel_to_data: dict, mapping from native_channel_name to spike data in microvolts
    """
    # Extract the number of channels
    num_channels = len(amplifier_channels)

    # Get file size in bytes
    fileinfo = os.path.getsize(file_path)

    # Calculate the number of samples
    num_samples = fileinfo // (num_channels * 2)  # int16 = 2 bytes

    # Initialize an empty dictionary to store the mapping
    channel_to_data = {}

    # Create a memory-mapped array that maps the binary file to an array
    fp = np.memmap(file_path, dtype='int16', mode='r', shape=(num_samples, num_channels))

    # Convert the data to microvolts and populate the dictionary
    for i, channel_info in enumerate(amplifier_channels):
        print(f"Reading channel {i}...")
        native_channel_name = channel_info.get("native_channel_name", f"Channel_{i}")
        channel_data = fp[:, i] * 0.195  # Conversion to microvolts
        channel_to_data[Channel(native_channel_name)] = channel_data

    return channel_to_data


def read_amplifier_data(file_path, amplifier_channels):
    """
    Reads amplifier data from a binary file and converts it to microvolts.
    Creates a dictionary mapping each native_channel_name to its vector of spikes.

    Parameters:
    - file_path: str, path to the amplifier.dat file
    - amplifier_channels: list of dicts, amplifier channel information

    Returns:
    - channel_to_data: dict, mapping from native_channel_name to spike data in microvolts
    """
    # Extract the number of channels
    num_channels = len(amplifier_channels)

    # Get file size in bytes
    fileinfo = os.path.getsize(file_path)

    # Calculate the number of samples
    num_samples = fileinfo // (num_channels * 2)  # int16 = 2 bytes

    # Initialize an empty dictionary to store the mapping
    channel_to_data = {}

    # Open the file and read the data
    with open(file_path, 'rb') as fid:
        v = np.fromfile(fid, dtype=np.int16, count=num_channels * num_samples)

    # Reshape and transpose the array to get the proper dimensions
    v = v.reshape((num_samples, num_channels)).T

    # Convert to microvolts
    v = v * 0.195

    # Populate the dictionary
    for i, channel_info in enumerate(amplifier_channels):
        native_channel_name = channel_info.get("native_channel_name", f"Channel_{i}")
        channel_to_data[Channel(native_channel_name)] = v[i, :]

    return channel_to_data



import xmltodict

from clat.compile.task.cached_task_fields import CachedTaskField
from clat.compile.tstamp.cached_tstamp_fields import CachedDatabaseField
from clat.compile.tstamp.tstamp_field import DatabaseField
from clat.util.connection import Connection
from clat.util.time_util import When


class StimSpecIdField(CachedDatabaseField):
    def get(self, when: When) -> int:
        return self.get_stim_spec_id(self.conn, when)

    def get_name(self):
        return "Id"

    def get_stim_spec_id(self, when: When) -> int:
        try:
            self.conn.execute(
                "SELECT msg from BehMsg WHERE "
                "type = 'SlideOn' AND "
                "tstamp >= %s AND tstamp <= %s",
                params=(int(when.start), int(when.stop)))
            trial_msg_xml = self.conn.fetch_one()
            trial_msg_dict = xmltodict.parse(trial_msg_xml)
            taskId = int(trial_msg_dict['SlideEvent']['taskId'])

            self.conn.execute("SELECT stim_id from TaskToDo WHERE "
                              "task_id = %s",
                              params=(taskId,))
            stim_spec_id = self.conn.fetch_one()
        except:
            return "None"
        return stim_spec_id


class StimSpecDataField(StimSpecIdField):
    def get(self, when: When) -> dict:
        stim_spec_id = super().get(when)
        return get_stim_spec_data(self.conn, when, stim_spec_id)


def get_stim_spec_data(conn: Connection, when: When, stim_spec_id) -> dict:
    """Given a tstamp of trialStart and trialStop, finds the stimSpec Id from Trial Message and then reads data from
    StimSpec """

    conn.execute("SELECT data from StimSpec WHERE "
                 "id = %s",
                 params=(stim_spec_id,))

    stim_spec_data_xml = conn.fetch_one()
    stim_spec_data_dict = xmltodict.parse(stim_spec_data_xml)
    return stim_spec_data_dict


def get_stim_spec(conn: Connection, when: When, stim_spec_id: int) -> dict:
    conn.execute("SELECT spec from StimSpec WHERE "
                 "id = %s",
                 params=(stim_spec_id,))
    stim_spec_xml = conn.fetch_one()
    stim_spec_dict = xmltodict.parse(stim_spec_xml)
    return stim_spec_dict


def get_ga_name_from_stim_spec_id(conn, stim_spec_id) -> str:
    conn.execute("SELECT ga_name FROM TaskToDo t WHERE"
                 " stim_id = %s",
                 params=(stim_spec_id,))

    ga_name = conn.fetch_one()
    return ga_name


class NewGaNameField(StimSpecIdField):
    def get(self, when: When) -> str:
        stim_spec_id = self.get_cached_super(when, StimSpecIdField)
        return get_new_ga_name_from_stim_spec_id(self.conn, stim_spec_id)

    def get_name(self):
        return "GaType"


def get_new_ga_name_from_stim_spec_id(conn, stim_spec_id):
    conn.execute("SELECT ga_name FROM TaskToDo t WHERE"
                 " stim_id = %s",
                 params=(stim_spec_id,))

    ga_name = conn.fetch_one()
    return ga_name


class NewGaLineageField(StimSpecIdField):
    def get(self, when: When) -> str:
        stim_spec_id = self.get_cached_super(when, StimSpecIdField)
        return get_new_ga_lineage_from_stim_spec_id(self.conn, stim_spec_id)

    def get_name(self):
        return "Lineage"


class RegimeScoreField(NewGaLineageField):
    def get(self, when: When) -> str:
        lineage_id = self.get_cached_super(when, NewGaLineageField)
        return float(get_regime_score_from_lineage_id(self.conn, lineage_id))

    def get_name(self):
        return "RegimeScore"


def get_regime_score_from_lineage_id(conn, lineage_id):
    conn.execute("SELECT regime FROM LineageGaInfo WHERE lineage_id"
                 " = %s",
                 params=(lineage_id,))
    regime_score = conn.fetch_one()
    return regime_score


def get_new_ga_lineage_from_stim_spec_id(conn, stim_spec_id):
    conn.execute("SELECT lineage_id FROM StimGaInfo WHERE"
                 " stim_id = %s",
                 params=(stim_spec_id,))

    lineage = conn.fetch_one()
    return lineage


class GaNameField(StimSpecIdField):
    def get(self, when: When) -> str:
        stim_spec_id = super().get(when)
        return get_ga_name_from_stim_spec_id(self.conn, stim_spec_id)


class GaTypeField(GaNameField):
    def get(self, when: When):
        ga_name = super().get(when)
        return get_ga_type_from_ga_name(self.conn, ga_name)


class GaLineageField(GaNameField):
    def get(self, when: When):
        ga_name = super().get(when)
        return get_ga_lineage_from_ga_name(self.conn, ga_name)


def get_ga_type_from_ga_name(conn, ga_name: str):
    ga_type = ga_name.split("-")[0]
    return ga_type


def get_ga_lineage_from_ga_name(conn, ga_name: str):
    ga_lineage = ga_name.split("-")[1]
    return ga_lineage


class TaskIdField(CachedDatabaseField):
    def get(self, when: When) -> int:
        try:
            self.conn.execute(
                "SELECT msg from BehMsg WHERE "
                "type = 'SlideOn' AND "
                "tstamp >= %s AND tstamp <= %s",
                params=(int(when.start), int(when.stop)))
            trial_msg_xml = self.conn.fetch_one()
            trial_msg_dict = xmltodict.parse(trial_msg_xml)
            taskId = int(trial_msg_dict['SlideEvent']['taskId'])

            return taskId
        except:
            return "None"

    def get_name(self):
        return "TaskId"


class StimIdField(TaskIdField):
    def get(self, when: When) -> int:
        task_id = self.get_cached_super(when, TaskIdField)
        self.conn.execute("SELECT stim_id from TaskToDo WHERE "
                          "task_id = %s",
                          params=(task_id,))
        stim_spec_id = self.conn.fetch_one()
        return stim_spec_id

    def get_name(self):
        return "StimId"


class StimSpecField(StimIdField):
    def __init__(self, conn: Connection):
        super().__init__(conn)

    def get(self, when: When) -> str:
        # Execute the query to get the StimSpec based on task_id
        # Note: Replace the query with the appropriate one for your schema
        stim_id = self.get_cached_super(when, StimIdField)
        query = "SELECT spec FROM StimSpec WHERE id = %s"
        params = (stim_id,)
        self.conn.execute(query, params)
        stim_spec = self.conn.fetch_one()
        return stim_spec

    def get_name(self):
        return "StimSpec"



from __future__ import annotations

from collections import OrderedDict

import ast
import pandas as pd

from clat.compile.tstamp.tstamp_field import DatabaseField
from clat.util.connection import Connection
from clat.util.time_util import When


class CachedDatabaseField(DatabaseField):
    """
    A DatabaseField that caches its value in the database so it can be used for analysis without need to refetch data or store in
    serialized intermediate format.

    Override get_cached_value and cache_value to change the caching logic.
    """

    def __init__(self, conn: Connection,
                 ):
        self.conn = conn
        self.name = self.get_name()
        super().__init__(conn, self.get_name())
        self._ensure_cache_table_exists()

    def get_cached_super(self, when: When, super_type: type[CachedDatabaseField], *args, **kwargs):
        """
        Get the value of the superclass instance of the specified type, caching it if necessary.
        If the field is not a superclass of this, it may fail if the superclass does not have compatible
        constructor parameters.

        for *args and **kwargs, pass the same arguments as the superclass constructor (excluding conn)
        """
        # Dynamically get the superclass instance based on super_type
        # Dynamically create an instance of the specified superclass
        super_field = super_type(self.conn, *args, **kwargs)

        # Attempt to retrieve cached value
        cached_value = self._get_cached_value(super_field.get_name(), when)
        if cached_value is not None:
            return self.convert_from_string(cached_value)

        # Fetch data using the superclass's get method and cache it
        data = super_field.get(when)
        self._cache_value(super_field.get_name(), when, data)
        converted_value = self.convert_from_string(self._get_cached_value(super_field.get_name(), when))
        return converted_value

    def get_and_cache(self, name: str, when: When):
        cached_value = self._get_cached_value(name, when)
        if cached_value is not None:
            return self.convert_from_string(cached_value)

        data = self.get(when)
        self._cache_value(name, when, data)
        # return the cached value rather than raw value to ensure same data-types are returned for all calls
        cached_value = self._get_cached_value(name, when)
        return self.convert_from_string(cached_value)

    def convert_from_string(self, cached_value):
        try:
            return ast.literal_eval(cached_value)
        except ValueError:
            return cached_value
        except SyntaxError:
            return cached_value

    def _get_cached_value(self, name: str, when: When):
        # Implement the logic to query the TrialFieldCache table
        # to retrieve the cached value, if it exists and is still valid.
        query = "SELECT value FROM TrialFieldCache WHERE name = %s AND start = %s AND stop = %s;"
        self.conn.execute(query, params=(name, int(when.start), int(when.stop)))
        result = self.conn.fetch_all()
        return result[0][0] if result else None

    def _cache_value(self, name: str, when: When, value):
        value = str(value)
        # Implement the logic to insert or update the cached value
        # in the TrialFieldCache table.
        query = """
        INSERT INTO TrialFieldCache (name, start, stop, value) 
        VALUES (%s, %s, %s, %s)
        ON DUPLICATE KEY UPDATE value = %s;
        """
        self.conn.execute(query, params=(name, int(when.start), int(when.stop), value, value))

    def _ensure_cache_table_exists(self):
        """Ensures that the TrialFieldCache table exists in the database."""
        # Check if the table exists
        check_table_query = """
        SELECT COUNT(*)
        FROM information_schema.tables
        WHERE table_schema = DATABASE()
        AND table_name = 'TrialFieldCache';
        """
        self.conn.execute(check_table_query)
        result = self.conn.fetch_all()

        # Create the table if it doesn't exist
        if result[0][0] == 0:
            create_table_query = """
            CREATE TABLE `TrialFieldCache` (
              `name` varchar(255) NOT NULL,
              `start` bigint(20) NOT NULL,
              `stop` bigint(20) NOT NULL,
              `value` longtext,
              PRIMARY KEY (`name`,`start`,`stop`)
            ) ENGINE=InnoDB DEFAULT CHARSET=latin1
            """
            self.conn.execute(create_table_query)

    def get_name(self) -> str:
        raise NotImplementedError("Subclasses must implement get_name")

class CachedFieldList(list[CachedDatabaseField]):
    def get_df(self):
        df = pd.DataFrame(columns=self.get_names())
        return df

    def get_names(self):
        return [field.get_name() for field in self]

    def to_data(self, trial_tstamps: list[When]) -> pd.DataFrame:
        return self._get_data_from_trials(trial_tstamps)

    def _get_data_from_trials(self, trial_tstamps: list[When]) -> pd.DataFrame:
        data = []
        for i, when in enumerate(trial_tstamps):
            print("working on " + str(i) + " out of " + str(len(trial_tstamps)))
            field_values = []
            for field in self:
                try:
                    field_values.append(field.get_and_cache(field.name, when))
                except Exception as e:
                    print(f"Error fetching {field.name} at {when} for trial {i}: {e}")
                    field_values.append(None)
            field_values.insert(0, when)
            names = self.get_names()
            names.insert(0, "TrialStartStop")
            new_row = OrderedDict(zip(names, field_values))
            data.append(new_row)

        return pd.DataFrame(data)


"""
CREATE TABLE `TrialFieldCache` (
  `name` varchar(255) NOT NULL,
  `start` bigint(20) NOT NULL,
  `stop` bigint(20) NOT NULL,
  `value` longtext,
  PRIMARY KEY (`name`,`start`,`stop`)
) ENGINE=InnoDB DEFAULT CHARSET=latin1
"""



import numpy as np
import os


def read_analogin_file(filename, num_channels):
    """
    Reads ADC data from a file and converts it to volts.

    Args:
        filename (str): The path to the binary file containing ADC data.
        num_channels (int): The number of channels in the ADC data.

    Returns:
        np.ndarray: A 2D array containing the voltage data for each channel.
    """
    # Get the size of the file in bytes
    file_size = os.path.getsize(filename)

    # Calculate the number of samples per channel
    num_samples = file_size // (num_channels * 2)  # uint16 = 2 bytes

    # Read the binary data from the file
    with open(filename, 'rb') as f:
        data = np.fromfile(f, dtype=np.uint16, count=num_channels * num_samples)

    # Reshape the data into a 2D array [num_channels, num_samples]
    data = data.reshape((num_channels, num_samples))

    # Convert ADC values to volts
    volts = (data - 32768) * 0.0003125

    return volts



import xmltodict

from clat.compile.task.base_database_fields import StimSpecDataField
from clat.util.connection import Connection


class MatchStickField(StimSpecDataField):
    def __init__(self, conn: Connection, name: str = None):
        super().__init__(conn, name)

    def get_mstick_data(self, task_id: int) -> dict:
        stim_spec_data_xml = super().get(task_id)
        stim_spec_data = xmltodict.parse(stim_spec_data_xml)
        # Assuming the mstick_data is part of the stim_spec_data, extract it
        mstick_data = stim_spec_data["AllenMStickData"]
        return mstick_data


class ShaftField(MatchStickField):
    def get(self, task_id: int) -> list[dict]:
        mstick_data = self.get_mstick_data(task_id)
        shaft_data = mstick_data['shaftData']['ShaftData']
        return shaft_data


class TerminationField(MatchStickField):
    def get(self, task_id: int) -> list[dict]:
        mstick_data = self.get_mstick_data(task_id)
        termination_data = mstick_data['terminationData']['TerminationData']
        return termination_data


class JunctionField(MatchStickField):
    def get(self, task_id: int) -> list[dict]:
        mstick_data = self.get_mstick_data(task_id)
        junction_data = mstick_data['junctionData']['JunctionData']
        return junction_data



import time
import datetime




class When:
    def __init__(self, start, stop):
        self.start = start
        self.stop = stop

    def tuple(self):
        return (self.start, self.stop)

    def __str__(self):
        return "({},{})".format(self.start, self.stop)

    def __repr__(self):
        return self.__str__()


def today() -> When:
    today = to_unix(datetime.date.today())
    tomorrow = to_unix(datetime.date.today() + datetime.timedelta(days=1))
    when = When(today, tomorrow)
    return when


def from_x_days_ago(x):
    start = to_unix(datetime.date.today() - datetime.timedelta(days=x))
    stop = now()
    when = When(start, stop)
    return when


def from_date_to_now(year: int, month: int, day: int) -> When:
    start = to_unix(datetime.date(year, month, day))
    stop = now()
    when = When(start, stop)
    return when

def all():
    when = When(0, to_unix(datetime.date.fromisoformat('3022-01-01')))
    return when


def now():
    return round(time.time() * 1000000)


def to_unix(dt: datetime.datetime):
    return round(time.mktime(dt.timetuple()) * 1000000)


def on_date(year: int, month: int, day: int) -> When:
    """
    Returns a When object representing the start and end of the 24-hour period for the given date.

    :param year: Year component of the date.
    :param month: Month component of the date.
    :param day: Day component of the date.
    :return: When object representing the start and end of the given date.
    """
    date = datetime.date(year, month, day)
    start = to_unix(date)
    end = to_unix(date + datetime.timedelta(days=1))
    return When(start, end)


def on_date_and_time(year: int, month: int, day: int, start_time: str = None, end_time: str = None) -> When:
    """
    Returns a When object representing the period from the given start date and optional start time to an optional end time or the end of the day if no end time is provided.

    :param year: Year component of the date.
    :param month: Month component of the date.
    :param day: Day component of the date.
    :param start_time: Optional start time as a string in 'HH:MM:SS' format. Defaults to the beginning of the day if not provided.
    :param end_time: Optional end time as a string in 'HH:MM:SS' format. Defaults to the end of the day if not provided.
    :return: When object representing the start and potential end of the given date and time.
    """
    # Determine the start time
    if start_time:
        # Create the start datetime from the date and time components
        start_datetime = datetime.datetime(year, month, day) + datetime.timedelta(seconds=int(start_time[:2])*3600 + int(start_time[3:5])*60 + int(start_time[6:]))
    else:
        # Default to the beginning of the day (00:00:00)
        start_datetime = datetime.datetime(year, month, day)
    start = to_unix(start_datetime)

    # Determine the end time
    if end_time:
        # Create the end datetime from the date and time components
        end_datetime = datetime.datetime(year, month, day) + datetime.timedelta(seconds=int(end_time[:2])*3600 + int(end_time[3:5])*60 + int(end_time[6:]))
    else:
        # Default to the end of the day (23:59:59)
        end_datetime = datetime.datetime(year, month, day, 23, 59, 59)
    end = to_unix(end_datetime)

    return When(start, end)



# Testing the new function with a start time and default end time
test_start_time = "08:30:00"  # 8:30 AM
when_for_date_time = on_date_and_time(2022, 7, 4, test_start_time)
when_for_date_time



import os
import struct
from typing import Any

from clat.intan.channels import Channel


# Version 3.0, 11 February 2021

# Reads spike.dat files generated by Intan Technologies RHX data acqusition
# software.  Data are parsed and placed into variables within the Python workspace.
# Therefore, it is recommended to add either saving these variables to disk or
# any plotting or processing of the data at the end of readIntanSpikeFile, before
# those variables are removed when execution completes

# Spike data from N channels are loaded into an N x M list named
# 'spikes', where M = 5 if spike snapshots were saved, otherwise M = 4.
# The first column of spikes contains native channel names.  The second
# column contains custom channel names.  The third column contains spike
# timestamps.  The fourth column contains spike ID numbers (128 = likely
# artifact due to amplitude exceeding threshold set in Spike Scope).  All 
# normal spikes have a spike ID of 1.  Future versions of the RHX software
# may support realtime spike sorting, and in this case spike ID will 
# denote different identified spikes (1, 2, 3, etc.).  If spike snapshots
# were saved then they are contained in the fifth column.

def read_intan_spike_file(full_file_name, no_artifacts=True) -> tuple[list[list[Any]], float]:
    """
    # Spike data from N channels are loaded into an N x M list named
    # 'spikes', where M = 5 if spike snapshots were saved, otherwise M = 4.
    # The first column of spikes contains native channel names.  The second
    # column contains custom channel names.  The third column contains spike
    # timestamps.  The fourth column contains spike ID numbers (128 = likely
    # artifact due to amplitude exceeding threshold set in Spike Scope).  All
    # normal spikes have a spike ID of 1.  Future versions of the RHX software
    # may support realtime spike sorting, and in this case spike ID will
    # denote different identified spikes (1, 2, 3, etc.).  If spike snapshots
    # were saved then they are contained in the fifth column.
    """

    # Open data file
    fid = open(full_file_name, 'rb')
    filesize = os.path.getsize(full_file_name)

    # Check 'magic number' at beginning of file to make sure this is an Intan
    # Technologies spike data file.
    magicNumber, = struct.unpack('<I', fid.read(4))
    if magicNumber == int('18f8474b', 16):
        multichannel = 1
    elif magicNumber == int('18f88c00', 16):
        multichannel = 0
    else:
        raise Exception('Unrecognized file type.')

    spikeFileVersionNumber, = struct.unpack('<H', fid.read(2))

    if spikeFileVersionNumber > 1:
        print("Warning: This spike file version is not supported by this file reader.")
        print("Check the Intan Technologies website for a more recent version.")

    filename = readString(fid)
    channelList = readString(fid).split(",")
    customChannelList = readString(fid).split(",")

    sample_rate, = struct.unpack('<f', fid.read(4))

    samplesPreDetect, = struct.unpack('<I', fid.read(4))
    samplesPostDetect, = struct.unpack('<I', fid.read(4))
    nSamples = samplesPreDetect + samplesPostDetect

    if nSamples == 0:
        snapshotsPresent = 0
    else:
        snapshotsPresent = 1

    N = len(channelList)

    spikes = [[] for _ in range(N)]
    for i in range(N):
        spikes[i].append(channelList[i])  # 0: native channel name
        spikes[i].append(customChannelList[i])  # 1: custom channel name
        spikes[i].append([])  # 2: single-float timestamp
        spikes[i].append([])  # 3: uint8 spike ID
        if snapshotsPresent:
            spikes[i].append([])  # 4: single-float snapshot

    while filesize - fid.tell() > 0:
        if multichannel:
            channelName = ""
            for charIndex in range(5):
                thisChar, = struct.unpack('<c', fid.read(1))
                channelName = channelName + str(thisChar, "utf-8")
            for i in range(N):
                if spikes[i][0] == channelName:
                    index = i
                    break
        else:
            index = 1

        timestamp, = struct.unpack('<i', fid.read(4))
        spikeID, = struct.unpack('<B', fid.read(1))

        if snapshotsPresent:
            snapshot = list(struct.unpack("<%dH" % nSamples, fid.read(2 * nSamples)))

        if spikeID == 128 and no_artifacts:
            continue

        timestampSeconds = timestamp / sample_rate

        spikes[index][2].append(timestampSeconds)
        spikes[index][3].append(spikeID)
        if snapshotsPresent:
            snapshotMicroVolts = [0.195 * (float(snapshotSample) - 32768.0) for snapshotSample in snapshot]
            spikes[i][4].append(snapshotMicroVolts)

    # Close data file
    fid.close()

    return spikes, sample_rate

    #
    # if snapshotsPresent:
    #     tSnapshot = [(sample - samplesPreDetect) / sampleRate for sample in range(nSamples)]
    #
    # # Just for demonstration, take the plot the 2nd (N = 1) channel's list of snapshots (snapshots are always
    # # in the fifth column M = 4). Grab the 6th snapshot present (list index = 5) and plot it
    # secondChannelSnapshots = spikes[1][4]
    # plt.plot(tSnapshot, secondChannelSnapshots[5])
    # plt.show()


def readString(fid):
    resultStr = ""
    ch, = struct.unpack('<c', fid.read(1))
    while ch != b'\0':
        resultStr = resultStr + str(ch, "utf-8")
        ch, = struct.unpack('<c', fid.read(1))
    return resultStr


# If the function is called with the "noartifacts" parameter, all spikes with spike ID = 128 are ignored.
# readIntanSpikeFile("artifacts")
# #readIntanSpikeFile("noartifacts")
def spike_matrix_to_spike_tstamps_for_channels(spike_matrix) -> dict[Channel: list[float]]:
    """
    Convert spike data into a dictionary of channel names and responses.
    """
    spike_dict = {}
    for row in spike_matrix:
        channel_name = str_to_channel_enum(row[0])
        responses = row[2]
        spike_dict[channel_name] = responses
    return spike_dict


def str_to_channel_enum(str_value):
    enum_name = str_value.replace('-', '_')
    try:
        return getattr(Channel, enum_name)
    except AttributeError:
        raise ValueError(f"{str_value} is not a valid member of {Channel.__name__}")


def fetch_spike_tstamps_from_file(spike_file_path: str) -> tuple[Any, float]:
    spike_matrix, sample_rate = read_intan_spike_file(spike_file_path)
    spike_tstamps_for_channels = spike_matrix_to_spike_tstamps_for_channels(spike_matrix)
    return spike_tstamps_for_channels, sample_rate



from clat.compile.task.task_field import TaskField
from clat.util.connection import Connection


class TaskIdField(TaskField):
    def __init__(self, name: str = "TaskId"):
        super().__init__(name)

    def get(self, task_id: int) -> int:
        return task_id


class DatabaseField(TaskField):
    def __init__(self, conn: Connection, name: str = None):
        super().__init__(name)
        self.conn = conn  # Assuming that the database connection is passed to the field


class StimSpecIdField(DatabaseField):
    def __init__(self, conn: Connection, name: str = "StimSpecId"):
        super().__init__(conn, name)

    def get(self, task_id: int) -> int:
        # Assuming a method or query to fetch the StimSpecId based on task_id
        # For demonstration, let's assume a method get_stim_spec_id is defined in the Connection class
        stim_spec_id = self.get_stim_spec_id(task_id)
        return int(stim_spec_id)

    def get_stim_spec_id(self, task_id: int) -> int:
        # Execute the query to get the StimSpecId based on task_id
        # Note: Replace the query with the appropriate one for your schema
        query = "SELECT stim_id FROM TaskToDo WHERE task_id = %s"
        params = (task_id,)
        self.conn.execute(query, params)
        stim_spec_id = self.conn.fetch_one()
        return stim_spec_id


class StimSpecField(StimSpecIdField):
    def __init__(self, conn: Connection, name: str = "StimSpec"):
        super().__init__(conn, name)

    def get(self, task_id: int) -> str:
        # Execute the query to get the StimSpec based on task_id
        # Note: Replace the query with the appropriate one for your schema
        stim_id = super().get(task_id)
        query = "SELECT spec FROM StimSpec WHERE id = %s"
        params = (stim_id,)
        self.conn.execute(query, params)
        stim_spec = self.conn.fetch_one()
        return stim_spec


class StimSpecDataField(StimSpecIdField):
    def __init__(self, conn: Connection, name: str = "StimSpecData"):
        super().__init__(conn, name)

    def get(self, task_id: int):
        # Execute the query to get the StimSpecData based on task_id
        # Note: Replace the query with the appropriate one for your schema
        stim_id = super().get(task_id)
        query = "SELECT data FROM StimSpec WHERE id = %s"
        params = (stim_id,)
        self.conn.execute(query, params)
        stim_spec_data = self.conn.fetch_one()
        return stim_spec_data



from matplotlib import pyplot as plt

from clat.intan.analogin import read_analogin_file


def test_read_analogin_file():
    volts = read_analogin_file("/run/user/1003/gvfs/sftp:host=172.30.9.78/home/i2_allen/Documents/Test/2024-02-28/TestRecording_240228_152822/analogin.dat", 1)
    # Plotting the voltages
    plt.figure(figsize=(10, 6))  # Set the figure size
    plt.plot(volts[0], label='Channel 1 Voltage')  # Plot the voltages for channel 1
    plt.title('Voltage Readings from analogin.dat')
    plt.xlabel('Sample Number')
    plt.ylabel('Voltage (V)')
    plt.legend()
    plt.grid(True)  # Add grid for better readability
    plt.show()






from tkinter import filedialog

from clat.eyecal.params import EyeCalibrationParameters
from clat.util.connection import Connection


def main():
    current_conn = Connection("allen_estimshape_train_231211")

    #Open a GUI window to select a file
    filename = filedialog.askopenfilename(initialdir = "/home/r2_allen/git/EStimShape/xper-train",title = "Select file")

    #Read the file as one big string
    with open(filename, 'r') as f:
        lines = f.readlines()

    string = ""
    for line in lines:
        string += line

    EyeCalibrationParameters.deserialize(string).write_params(current_conn)




if __name__ == '__main__':
    main()



import bisect
import os
from dataclasses import dataclass

from clat.intan.livenotes import map_task_id_to_epochs_with_livenotes
from clat.intan.marker_channels import epoch_using_marker_channels
from clat.intan.spike_file import fetch_spike_tstamps_from_file


@dataclass
class OneFileParser:
    import os
    import bisect
    sample_rate: int = None

    def parse(self, intan_file_path: str) -> tuple[
        dict[int, dict[int, list[float]]], dict[int, tuple[float, float]], int]:
        """
        Returns a dictionary of epoched spikes by channel by task_id,
        a dictionary of the epoch start and stop times by task_id,
        and the sample rate.
        filtered_spikes_by_channel_by_task_id: Dict[taskId, Dict[Channel, Responses]].
        epoch_start_stop_times_by_task_id: Dict[taskId, Tuple[StartSeconds, EndSeconds]].
        sample_rate: int.
        """
        spike_path = os.path.join(intan_file_path, "spike.dat")
        digital_in_path = os.path.join(intan_file_path, "digitalin.dat")
        notes_path = os.path.join(intan_file_path, "notes.txt")

        spike_tstamps_by_channel, self.sample_rate = fetch_spike_tstamps_from_file(spike_path)
        stim_epochs_from_markers = epoch_using_marker_channels(digital_in_path, false_negative_correction_duration=2)
        epochs_for_task_ids = map_task_id_to_epochs_with_livenotes(notes_path, stim_epochs_from_markers,
                                                                   require_trial_complete=False)

        filtered_spikes_by_channel_by_task_id = {}
        epoch_start_stop_times_by_task_id = {}

        # Ensure all timestamps are sorted if not already sorted
        for channel, tstamps in spike_tstamps_by_channel.items():
            spike_tstamps_by_channel[channel] = sorted(tstamps)

        for task_id, epoch_indices in epochs_for_task_ids.items():
            print(f"Epoching task_id: {task_id}")
            filtered_spikes_for_channels = {}

            for channel, tstamps in spike_tstamps_by_channel.items():
                # Using binary search to find the range of timestamps within the current epoch
                start_index = bisect.bisect_left(tstamps, epoch_indices[0] / self.sample_rate)
                end_index = bisect.bisect_right(tstamps, epoch_indices[1] / self.sample_rate)
                # Extract the timestamps that fall within the epoch
                passed_filter = tstamps[start_index:end_index]
                filtered_spikes_for_channels[channel] = passed_filter

            epoch_start_seconds = epoch_indices[0] / self.sample_rate
            epoch_end_seconds = epoch_indices[1] / self.sample_rate
            epoch_start_stop_times_by_task_id[task_id] = (epoch_start_seconds, epoch_end_seconds)
            filtered_spikes_by_channel_by_task_id[task_id] = filtered_spikes_for_channels

        return filtered_spikes_by_channel_by_task_id, epoch_start_stop_times_by_task_id, self.sample_rate



from __future__ import annotations

from typing import Union

"""This module contains functions for working with nested dictionaries and lists of (nested) dictionaries """


def apply_function_to_subdictionaries_values_with_keys(dictionary: Union[dict, list], keys: list, function):
    """apply function to the subdictionary that contains the specified keys
    The supplied function should take in the dictionary containing the specified keys and
    return a new dictionary"""
    if isinstance(dictionary, dict):
        if set(keys).issubset(dictionary.keys()):
            dictionary = function(dictionary)
        else:
            for key, value in dictionary.items():
                dictionary[key] = apply_function_to_subdictionaries_values_with_keys(value, keys, function)
    elif isinstance(dictionary, list):
        for index, item in enumerate(dictionary):
            dictionary[index] = apply_function_to_subdictionaries_values_with_keys(item, keys, function)
    return dictionary


def check_condition_on_subdictionaries(dictionary: Union[dict, list], condition, boolean_to_update, *args):
    """Returns true if any of the key-value pairs satisfy the condition. The condition is a function that returns a boolean
    and has parameters: (key, value, *args)"""
    if isinstance(dictionary, dict):
        for key, value in dictionary.items():
            if isinstance(value, dict):
                if check_condition_on_subdictionaries(value, condition, boolean_to_update, *args):
                    return True
            else:
                if condition(key, value, *args):
                    return True

    elif isinstance(dictionary, list):
        for item in dictionary:
            if check_condition_on_subdictionaries(item, condition, boolean_to_update, *args):
                return True


def flatten_dictionary(dictionary: Union[dict, list], output_value_list: list, output_key_list):
    """Converts a nested dictionary into a single list of values and appends those values into output_list
    :param output_key_list:
    """
    if isinstance(dictionary, dict):
        for key, value in dictionary.items():
            if isinstance(value, dict):
                flatten_dictionary(value, output_value_list, output_key_list)
            else:
                output_value_list.append(value)
                if output_key_list is not None:
                    output_key_list.append(key)

    elif isinstance(dictionary, list):
        for value in dictionary:
            if isinstance(value, dict) or isinstance(value, list):
                flatten_dictionary(value, output_value_list, output_key_list)
            else:
                output_value_list.append(value)
                if output_key_list is not None:
                    output_key_list.append(None)




def extract_values_with_key_into_list(dictionary: Union[dict, list], output_list: list, key: str):
    if isinstance(dictionary, dict):
        for k, value in dictionary.items():
            if isinstance(value, dict):
                extract_values_with_key_into_list(value, output_list, key)
            else:
                if k == key:
                    output_list.append(value)

    elif isinstance(dictionary, list):
        for value in dictionary:
            if isinstance(value, dict) or isinstance(value, list):
                extract_values_with_key_into_list(value, output_list, key)
            else:
                output_list.append(value)



#! /bin/env python
#
# Michael Gibson 23 April 2015
# Modified Adrian Foy Sep 2018

import sys, struct
import numpy as np

def read_one_data_block(data, header, indices, fid):
    """Reads one 60 or 128 sample data block from fid into data, at the location indicated by indices."""

    # In version 1.2, we moved from saving timestamps as unsigned
    # integers to signed integers to accommodate negative (adjusted)
    # timestamps for pretrigger data['
    if (header['version']['major'] == 1 and header['version']['minor'] >= 2) or (header['version']['major'] > 1):
        data['t_amplifier'][indices['amplifier']:(indices['amplifier'] + header['num_samples_per_data_block'])] = np.array(struct.unpack('<' + 'i' * header['num_samples_per_data_block'], fid.read(4 * header['num_samples_per_data_block'])))
    else:
        data['t_amplifier'][indices['amplifier']:(indices['amplifier'] + header['num_samples_per_data_block'])] = np.array(struct.unpack('<' + 'I' * header['num_samples_per_data_block'], fid.read(4 * header['num_samples_per_data_block'])))

    if header['num_amplifier_channels'] > 0:
        tmp = np.fromfile(fid, dtype='uint16', count= header['num_samples_per_data_block'] * header['num_amplifier_channels'])
        data['amplifier_data'][range(header['num_amplifier_channels']), (indices['amplifier']):(indices['amplifier']+ header['num_samples_per_data_block'])] = tmp.reshape(header['num_amplifier_channels'], header['num_samples_per_data_block'])

    if header['num_aux_input_channels'] > 0:
        tmp = np.fromfile(fid, dtype='uint16', count= int((header['num_samples_per_data_block'] / 4) * header['num_aux_input_channels']))
        data['aux_input_data'][range(header['num_aux_input_channels']), indices['aux_input']:int(indices['aux_input']+ (header['num_samples_per_data_block']/4))] = tmp.reshape(header['num_aux_input_channels'], int(header['num_samples_per_data_block']/4))

    if header['num_supply_voltage_channels'] > 0:
        tmp = np.fromfile(fid, dtype='uint16', count=1 * header['num_supply_voltage_channels'])
        data['supply_voltage_data'][range(header['num_supply_voltage_channels']), indices['supply_voltage']:(indices['supply_voltage']+1)] = tmp.reshape(header['num_supply_voltage_channels'], 1)

    if header['num_temp_sensor_channels'] > 0:
        tmp = np.fromfile(fid, dtype='uint16', count=1 * header['num_temp_sensor_channels'])
        data['temp_sensor_data'][range(header['num_temp_sensor_channels']), indices['supply_voltage']:(indices['supply_voltage']+1)] = tmp.reshape(header['num_temp_sensor_channels'], 1)

    if header['num_board_adc_channels'] > 0:
        tmp = np.fromfile(fid, dtype='uint16', count= (header['num_samples_per_data_block']) * header['num_board_adc_channels'])
        data['board_adc_data'][range(header['num_board_adc_channels']), indices['board_adc']:(indices['board_adc']+ header['num_samples_per_data_block'])] = tmp.reshape(header['num_board_adc_channels'], header['num_samples_per_data_block'])

    if header['num_board_dig_in_channels'] > 0:
        data['board_dig_in_raw'][indices['board_dig_in']:(indices['board_dig_in']+ header['num_samples_per_data_block'])] = np.array(struct.unpack('<' + 'H' * header['num_samples_per_data_block'], fid.read(2 * header['num_samples_per_data_block'])))

    if header['num_board_dig_out_channels'] > 0:
        data['board_dig_out_raw'][indices['board_dig_out']:(indices['board_dig_out']+ header['num_samples_per_data_block'])] = np.array(struct.unpack('<' + 'H' * header['num_samples_per_data_block'], fid.read(2 * header['num_samples_per_data_block'])))




from __future__ import annotations

import ast
from concurrent.futures import ThreadPoolExecutor
from tkinter import filedialog
import tkinter as tk
import numpy as np
import xmltodict

from clat.compile.tstamp.cached_tstamp_fields import CachedDatabaseField, CachedFieldList
from clat.eyecal.params import EyeCalibrationParameters
from clat.util import time_util

from clat.compile.tstamp.trial_tstamp_collector import TrialCollector
from clat.compile.tstamp.tstamp_field import DatabaseField, FieldList, get_data_from_trials
from clat.util.connection import Connection
from clat.util.time_util import When
from typing import Optional, Tuple, Any, List
from datetime import datetime
import matplotlib.pyplot as plt


def main():
    current_conn = Connection("allen_estimshape_ga_train_240604")
    trial_collector = TrialCollector(conn=current_conn, when=time_util.from_x_days_ago(0))
    calibration_trial_times = trial_collector.collect_calibration_trials()
    calibration_trial_times = filter_messages_after_experiment_start(current_conn, calibration_trial_times)
    print("calibration_trial_times: " + str(calibration_trial_times))

    fields = CachedFieldList()
    fields.append(CalibrationPointPositionField(current_conn))
    fields.append(SlideOnOffTimestampField(current_conn))
    fields.append(AverageVoltsField(current_conn))
    fields.append(DegreesField(current_conn))
    data = fields.to_data(calibration_trial_times)

    plot_average_volts(data)

    user_response = input("Do you want to serialize the current parameters? (yes/no): ").strip().lower()
    if user_response == 'yes':
        # Get the current parameters
        params = EyeCalibrationParameters.read_params(current_conn)
        serialized_params = params.serialize()
        print(serialized_params)

        # Open a GUI window to select file save location
        root = tk.Tk()
        root.withdraw()  # Hide the main window
        file_path = filedialog.asksaveasfilename(defaultextension=".txt", filetypes=[("Text files", "*.txt")])

        # Save the serialized data to a file, if a file was selected
        if file_path:
            with open(file_path, 'w') as file:
                file.write(serialized_params)

        root.destroy()


def filter_messages_after_experiment_start(conn, calibration_trial_times):
    # Find the timestamp of the most recent "ExperimentStart" message
    query = """
        SELECT MAX(tstamp) 
        FROM BehMsg 
        WHERE type = 'ExperimentStart'
    """
    conn.execute(query)
    experiment_start_timestamp = conn.fetch_all()[0][0]

    filtered_messages = []

    # Loop through each trial time range in calibration_trial_times
    for when in calibration_trial_times:
        # Ensure the trial is after the experiment start
        if when.start > experiment_start_timestamp:
            filtered_messages.append(when)

    return filtered_messages


def hash_tuple(t):
    """ Hash a tuple to a unique value """
    return hash(t)


def plot_average_volts(data):
    # Define five distinct colors
    colors = ['red', 'green', 'blue', 'yellow', 'purple']

    # Extracting data for average volts and degrees
    left_eye_avg_volts = data['AverageVoltsLeftRight'].apply(lambda x: x[0])
    right_eye_avg_volts = data['AverageVoltsLeftRight'].apply(lambda x: x[1])
    degrees_left_right = data['DegreesLeftRight']
    calibration_points = data['CalibrationPointPosition']

    # Assign colors to each unique calibration point
    unique_points = sorted(set(calibration_points))
    color_mapping = {point: colors[i] for i, point in enumerate(unique_points)}
    calibration_colors = calibration_points.map(color_mapping)

    # Create 2x2 subplots
    fig, axs = plt.subplots(2, 2, figsize=(15, 12))

    # Plot for left eye volts
    axs[0, 0].scatter([x[0] for x in left_eye_avg_volts], [y[1] for y in left_eye_avg_volts], c=calibration_colors)
    axs[0, 0].set_title('Average Volt Positions - Left Eye')
    axs[0, 0].set_xlabel('X Position')
    axs[0, 0].set_ylabel('Y Position')

    # Plot for right eye volts
    axs[0, 1].scatter([x[0] for x in right_eye_avg_volts], [y[1] for y in right_eye_avg_volts], c=calibration_colors)
    axs[0, 1].set_title('Average Volt Positions - Right Eye')
    axs[0, 1].set_xlabel('X Position')
    axs[0, 1].set_ylabel('Y Position')

    # Plot for left eye degrees
    degrees_left = [degree[0] for degree in degrees_left_right]
    axs[1, 0].scatter([x[0] for x in degrees_left], [y[1] for y in degrees_left], c=calibration_colors)
    axs[1, 0].set_title('Degree Positions - Left Eye')
    axs[1, 0].set_xlabel('X Position')
    axs[1, 0].set_ylabel('Y Position')

    # Plot for right eye degrees
    degrees_right = [degree[1] for degree in degrees_left_right]
    axs[1, 1].scatter([x[0] for x in degrees_right], [y[1] for y in degrees_right], c=calibration_colors)
    axs[1, 1].set_title('Degree Positions - Right Eye')
    axs[1, 1].set_xlabel('X Position')
    axs[1, 1].set_ylabel('Y Position')

    # Show plot
    plt.tight_layout()
    plt.show()


class CalibrationPointPositionField(CachedDatabaseField):
    def __init__(self, conn):
        super().__init__(conn)

    def get_name(self):
        return "CalibrationPointPosition"

    def get(self, when: When):
        return self.get_calibration_point_setup_msg(when.start, when.stop)

    def get_calibration_point_setup_msg(self, start_tstamp: datetime, end_tstamp: datetime) -> tuple[Any, Any]:
        query = """
            SELECT msg 
            FROM BehMsg 
            WHERE type = 'CalibrationPointSetup' 
            AND tstamp BETWEEN %s AND %s
        """
        params = (start_tstamp, end_tstamp)
        self.conn.execute(query, params)
        result = self.conn.fetch_all()
        msg = result[0][0] if result else None
        msg_dict = xmltodict.parse(msg)
        x = msg_dict['CalibrationPointSetupMessage']['fixationPosition']['x']
        y = msg_dict['CalibrationPointSetupMessage']['fixationPosition']['y']
        return (x, y)


class SlideOnOffTimestampField(CachedDatabaseField):
    def __init__(self, conn):
        super().__init__(conn)

    def get_name(self):
        return "SlideOnOffTimestamps"

    def get(self, when: When) -> Tuple[Optional[Any], Optional[Any]]:
        return self.get_slide_on_off_timestamps(when.start, when.stop)

    def get_slide_on_off_timestamps(self, start_tstamp: datetime, end_tstamp: datetime) -> Tuple[
        Optional[Any], Optional[Any]]:
        slide_on_query = """
            SELECT tstamp 
            FROM BehMsg 
            WHERE type = 'SlideOn' 
            AND tstamp BETWEEN %s AND %s
            ORDER BY tstamp ASC
            LIMIT 1
        """
        slide_off_query = """
            SELECT tstamp 
            FROM BehMsg 
            WHERE type = 'SlideOff' 
            AND tstamp BETWEEN %s AND %s
            ORDER BY tstamp ASC
            LIMIT 1
        """

        # Execute the query for SlideOn
        params = (start_tstamp, end_tstamp)
        self.conn.execute(slide_on_query, params)
        result = self.conn.fetch_all()
        slide_on_timestamp = result[0][0] if result else None

        # Execute the query for SlideOff
        self.conn.execute(slide_off_query, params)
        result = self.conn.fetch_all()
        slide_off_timestamp = result[0][0] if result else None

        return slide_on_timestamp, slide_off_timestamp



class AverageVoltsField(SlideOnOffTimestampField):
    def __init__(self, conn):
        super().__init__(conn)

    def get_name(self):
        return "AverageVoltsLeftRight"

    def get(self, when: When) -> Tuple[Optional[Tuple[float, float]], Optional[Tuple[float, float]]]:
        slide_on_off_timestamps = self.get_cached_super(when, SlideOnOffTimestampField)
        left_eye_positions, right_eye_positions = self.get_eye_device_messages(slide_on_off_timestamps[0], slide_on_off_timestamps[1])
        left_eye_positions_filtered = self.remove_outliers(left_eye_positions)
        right_eye_positions_filtered = self.remove_outliers(right_eye_positions)

        average_left = self.calculate_average(left_eye_positions_filtered)
        average_right = self.calculate_average(right_eye_positions_filtered)

        # self._cache_value(self.get_name(), when, (average_left, average_right))
        return average_left, average_right

    def get_eye_device_messages(self, start_tstamp: datetime, end_tstamp: datetime) -> Tuple[
        List[Tuple[float, float]], List[Tuple[float, float]]]:


        query = """
            SELECT msg
                FROM BehMsgEye
                WHERE type = 'EyeDeviceMessage' 
                AND tstamp BETWEEN %s AND %s
        """
        params = (start_tstamp, end_tstamp)
        self.conn.execute(query, params)
        results = self.conn.fetch_all()

        # Process results in parallel
        with ThreadPoolExecutor() as executor:
            processed_results = list(executor.map(self.process_message, results))

        if not processed_results:
            return ([(0,0), (0,0)], [(0,0), (0,0)])
        left_eye_positions, right_eye_positions = zip(*processed_results)
        return list(filter(None, left_eye_positions)), list(filter(None, right_eye_positions))


    @staticmethod
    def process_message(row):
        msg = row[0]
        msg_dict = xmltodict.parse(msg)
        eye_id = msg_dict['EyeDeviceMessage']['id']
        volt_x = float(msg_dict['EyeDeviceMessage']['volt']['x'])
        volt_y = float(msg_dict['EyeDeviceMessage']['volt']['y'])

        if eye_id == 'leftIscan':
            return (volt_x, volt_y), None
        elif eye_id == 'rightIscan':
            return None, (volt_x, volt_y)
        return None, None
    @staticmethod
    def calculate_average(positions: List[Tuple[float, float]]) -> Optional[Tuple[float, float]]:
        if not positions:
            return None

        sum_x = sum(pos[0] for pos in positions)
        sum_y = sum(pos[1] for pos in positions)
        count = len(positions)

        return sum_x / count, sum_y / count

    @staticmethod
    def remove_outliers(positions: List[Tuple[float, float]]) -> List[Tuple[float, float]]:
        if not positions:
            return []

        # Flatten the list of tuples and separate into x and y components
        x_vals, y_vals = zip(*positions)
        x_vals_filtered = AverageVoltsField.filter_outlier_values(x_vals)
        y_vals_filtered = AverageVoltsField.filter_outlier_values(y_vals)

        # Reconstruct the list of tuples with filtered values
        return list(zip(x_vals_filtered, y_vals_filtered))

    @staticmethod
    def filter_outlier_values(values):
        data = np.array(values)
        median = np.percentile(data, 50)
        q1 = np.percentile(data, 25)
        q3 = np.percentile(data, 75)
        iqr = q3 - q1
        lower_bound = median - 1.5 * iqr
        upper_bound = median + 1.5 * iqr
        return data[(data >= lower_bound) & (data <= upper_bound)]


class DegreesField(AverageVoltsField):
    def get_name(self):
        return "DegreesLeftRight"

    def __init__(self, conn):
        super().__init__(conn)

    def get(self, when: When):
        params = EyeCalibrationParameters.read_params(self.conn)
        left_eye_positions, right_eye_positions = self.get_cached_super(when, AverageVoltsField)
        return params.volt_to_degree((left_eye_positions, right_eye_positions))


if __name__ == '__main__':
    main()



import numpy
import numpy as np
from clat.util import table_util
from clat.util import time_util
from clat.util.connection import Connection
from clat.util.time_util import When


class TrialCollector:
    """Class for collecting trials from the database based on trialStart and trialStop tstamps in BehMsg,
       and filtering types of trials (i.e. from different kinds of experiments)
       based on the msgs between trialStart and trialStop in BehMsg"""
    def __init__(self, conn: Connection, when: When = time_util.all()):
        self.conn = conn
        self.when = when
        # self.beh_msg = conn.get_beh_msg(when)
        # self.stim_spec = conn.get_stim_spec(when)
        # self.stim_obj_data = conn.get_stim_obj_data(when)

    def collect_trials(self):
        print("Collecting all trials")

        # SQL query to get all trial start timestamps
        query_trial_starts = """
        SELECT tstamp 
        FROM BehMsg 
        WHERE type = 'TrialStart'
        AND tstamp BETWEEN %s AND %s
        ORDER BY tstamp;
        """
        self.conn.execute(query_trial_starts, (self.when.start, self.when.stop))
        trial_starts = self.conn.fetch_all()
        trial_starts = [trial_start_tuple[0] for trial_start_tuple in trial_starts]

        # SQL query to get all trial stop timestamps
        query_trial_stops = """
        SELECT tstamp 
        FROM BehMsg 
        WHERE type = 'TrialStop'
        AND tstamp BETWEEN %s AND %s
        ORDER BY tstamp;
        """
        self.conn.execute(query_trial_stops, (self.when.start, self.when.stop))
        trial_stops = self.conn.fetch_all()
        trial_stops = [trial_stop_tuple[0] for trial_stop_tuple in trial_stops]

        trial_starts, trial_stops = self.sort_fix_bad_trials(trial_starts, trial_stops)
        return [time_util.When(trial_starts[i], trial_stops[i]) for i in
                range(min(len(trial_starts), len(trial_stops)))]

    def collect_choice_trials(self):
        all_trial_whens = self.collect_trials()
        print("Collecting choice trials")
        choice_trial_whens = []


        query = """
                SELECT tstamp
                FROM BehMsg 
                WHERE type = 'ChoiceSelectionSuccess' 
                AND tstamp BETWEEN %s AND %s;
                """
        self.conn.execute(query, (int(self.when.start), int(self.when.stop)))
        result = self.conn.fetch_all()
        choice_tstamps = [result_tuple[0] for result_tuple in result]

        # Check if any of the choice_tstamps are sandwiched between the trial start and stop times
        for when in all_trial_whens:
            for choice_tstamp in choice_tstamps:
                if when.start <= choice_tstamp <= when.stop:
                    choice_trial_whens.append(when)
                    break  # Once a choice_tstamp is found in a trial period, no need to check other choice_tstamps

        return choice_trial_whens

    def sort_fix_bad_trials(self, trial_starts, trial_stops):
        """Finds bad trials by sorting the trial start and stops by their time stamps
        in the same array. We should have start, stop, start, stop, ..., and so on.
        If there's a duplicate start (i.e start, start, stop): then we delete the earlier trial_start
        If there's a duplicate stop (i.e, start, stop, stop):  then we delete the later trial_stop
        """
        self.__ensure_ends_are_aligned(trial_starts, trial_stops)

        trial_starts_ids = ["trial_start" for trial_start in trial_starts]
        trial_stops_ids = ["trial_stop" for trial_stop in trial_stops]
        combined_ids = trial_starts_ids + trial_stops_ids
        combined_tstamps = numpy.append(trial_starts, trial_stops)
        combined = zip(combined_ids, combined_tstamps)
        combined_sorted = sorted(combined, key=lambda x: x[1])

        previous = 'trial_stop'
        to_remove = []
        for i, item in enumerate(combined_sorted):
            if item[0] == previous:
                if item[0] == 'trial_start':
                    to_remove.append(i - 1)  # remove earlier trial_start, since it's missing a trial_stop
                    print("removing duplicate trial_start")
                else:
                    print("Somehow we have two repeating stops!: " + str(i))
                    to_remove.append(i)
            previous = item[0]
        combined_sorted_removed = [item for i, item in enumerate(combined_sorted) if i in to_remove]
        combined_sorted = [item for i, item in enumerate(combined_sorted) if i not in to_remove]
        trial_starts = [item[1] for item in combined_sorted if item[0] == 'trial_start']
        trial_stops = [item[1] for item in combined_sorted if item[0] == 'trial_stop']
        return trial_starts, trial_stops

    def collect_calibration_trials(self):
        all_trial_times = self.collect_trials()
        calibration_trial_times = []

        for when in all_trial_times:
            # SQL query to check if there are any 'CalibrationPointSetup' messages during the trial
            query = """
                    SELECT COUNT(*) 
                    FROM BehMsg 
                    WHERE type = 'CalibrationPointSetup' 
                    AND tstamp BETWEEN %s AND %s;
                    """
            self.conn.execute(query, (int(when.start), int(when.stop)))
            result = self.conn.fetch_all()

            # If there are any 'CalibrationPointSetup' messages, add this trial to the list
            if result[0][0] > 0:
                when.start = int(when.start)
                when.stop = int(when.stop)
                calibration_trial_times.append(when)

        return calibration_trial_times

    def __ensure_ends_are_aligned(self, trial_starts, trial_stops):
        while trial_stops[0] < trial_starts[0]:
            trial_stops = trial_stops[1:]
        while trial_starts[-1] > trial_stops[-1]:
            trial_starts = trial_starts[:-1]
        return trial_starts, trial_stops

    def __ensure_balanced_trial_nums(self, trial_starts, trial_stops):
        while trial_starts.size != trial_stops.size:
            if trial_starts.size > trial_stops.size:
                diff_length = trial_starts.size - trial_stops.size
                try:
                    first_bad_trial = self.__get_first_bad_trial(trial_starts[:-diff_length], trial_stops)
                except:
                    first_bad_trial = self.__get_first_bad_trial(trial_starts[diff_length:], trial_stops)
                trial_starts = np.delete(trial_starts, first_bad_trial)
            else:
                diff_length = trial_stops.size - trial_starts.size
                try:
                    first_bad_trial = self.__get_first_bad_trial(trial_starts, trial_stops[:-diff_length])
                except:
                    first_bad_trial = self.__get_first_bad_trial(trial_starts, trial_stops[diff_length:])
                trial_stops = np.delete(trial_stops, first_bad_trial)
        return trial_starts, trial_stops

    def __remove_misaligned_trials(self, trial_starts, trial_stops):
        while not self.__trials_aligned(trial_starts, trial_stops):
            first_bad_trial = self.__get_first_bad_trial(trial_starts, trial_stops)
            trial_starts = np.delete(trial_starts, first_bad_trial)
            trial_stops = np.delete(trial_stops, first_bad_trial)
        return trial_starts, trial_stops

    def __get_first_bad_trial(self, trial_starts, trial_stops):
        bad_trials = np.array(trial_starts > trial_stops)
        first_bad_trial = [i for i, x in enumerate(bad_trials) if x][0]
        return first_bad_trial

    def __trials_aligned(self, trial_starts, trial_stops):
        actual_correctly_aligned = sum([True for i in range(len(trial_starts)) if (trial_starts[i] < trial_stops[i])])
        return actual_correctly_aligned == len(trial_starts)


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    pass



from concurrent.futures import ProcessPoolExecutor

import logging
import pandas as pd
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Generic, TypeVar, Dict, Any, List, Tuple, Union

# Type variables for input/output types
InputT = TypeVar('InputT')
OutputT = TypeVar('OutputT')
ResultT = TypeVar('ResultT')

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger('neurophys.pipeline')


###################
# CORE INTERFACES #
###################

class InputHandler(Generic[InputT, OutputT], ABC):
    """Base class for input handlers that prepare data for computation."""

    @abstractmethod
    def prepare(self, data: InputT) -> OutputT:
        """Transform input data into format needed by computation module."""
        pass


class ComputationModule(Generic[InputT, OutputT], ABC):
    """Base class for analysis computation modules."""

    @abstractmethod
    def compute(self, prepared_data: InputT) -> OutputT:
        """Perform the core analysis computation."""
        pass

    @property
    def metadata(self) -> Dict[str, Any]:
        """Return metadata about the computation."""
        return {
            'name': self.__class__.__name__,
            'parameters': self.get_parameters()
        }

    def get_parameters(self) -> Dict[str, Any]:
        """Return computation parameters."""
        # Get all instance attributes that don't start with underscore
        return {
            key: value for key, value in self.__dict__.items()
            if not key.startswith('_')
        }


class OutputHandler(Generic[InputT, ResultT], ABC):
    """Base class for output handlers that format or visualize results."""

    @abstractmethod
    def process(self, result: InputT) -> ResultT:
        """Process the computation results."""
        pass


class AnalysisModule(Generic[InputT, OutputT, ResultT]):
    """Combined module for a complete analysis step."""

    def __init__(self,
                 input_handler: InputHandler,
                 computation: ComputationModule,
                 output_handler: OutputHandler,
                 name: str = None):
        self.input_handler = input_handler
        self.computation = computation
        self.output_handler = output_handler
        self.name = name or self.__class__.__name__
        self._results = None
        self._raw_output = None

    def run(self, data: InputT) -> ResultT:
        """Execute the full analysis module."""
        try:
            logger.info(f"Module {self.name}: Starting preparation...")
            prepared_data = self.input_handler.prepare(data)

            logger.info(f"Module {self.name}: Running computation...")
            computation_result = self.computation.compute(prepared_data)
            self._raw_output = computation_result

            logger.info(f"Module {self.name}: Processing output...")
            output = self.output_handler.process(computation_result)
            self._results = output

            return output
        except Exception as e:
            logger.error(f"Error in module {self.name}: {str(e)}")
            raise

    @property
    def results(self):
        """Access the processed results."""
        return self._results

    @property
    def raw_output(self):
        """Access the raw computation results before output processing."""
        return self._raw_output

    def __str__(self):
        return f"AnalysisModule({self.name})"

    def __repr__(self):
        return self.__str__()


class MultiInputHandler(InputHandler):
    """Handles multiple inputs from parallel branches."""

    def __init__(self, handlers=None):
        """Initialize with optional specific handlers for each input."""
        self.handlers = handlers or []

    def prepare(self, data_tuple):
        """Process each element in the tuple with corresponding handler."""
        if not isinstance(data_tuple, tuple):
            raise TypeError("Expected a tuple of inputs")

        # If specific handlers are provided, use them
        if self.handlers:
            if len(self.handlers) != len(data_tuple):
                raise ValueError(f"Got {len(data_tuple)} inputs but have {len(self.handlers)} handlers")

            prepared_data = []
            for i, (handler, data) in enumerate(zip(self.handlers, data_tuple)):
                prepared_data.append(handler.prepare(data))
            return tuple(prepared_data)

        # Default behavior: pass through all inputs
        return data_tuple


class PassthroughHandler(InputHandler):
    """Simple handler that passes data through unchanged."""

    def prepare(self, data):
        """Return the data unchanged."""
        return data


class DataFrameOutputHandler(OutputHandler):
    """Standard output handler that ensures results are in DataFrame format."""

    def process(self, result):
        """Convert result to DataFrame if needed."""
        if isinstance(result, pd.DataFrame):
            return result
        elif isinstance(result, dict):
            return pd.DataFrame([result])
        elif isinstance(result, list) and all(isinstance(item, dict) for item in result):
            return pd.DataFrame(result)
        else:
            # Just return as is if we don't know how to convert
            return result


####################
# PIPELINE CLASSES #
####################

# Type definition for pipeline structures
StructureType = Union[
    AnalysisModule,
    List['StructureType'],
    Tuple['StructureType', ...],
    Dict[str, Any]
]


@dataclass
class PipelineResult:
    """Container for pipeline execution results."""
    final_output: Any
    step_outputs: Dict[str, Any] = field(default_factory=dict)
    execution_time: Dict[str, float] = field(default_factory=dict)

    def get_step_result(self, step_name: str) -> Any:
        """Get the result of a specific named step."""
        return self.step_outputs.get(step_name)


class AnalysisPipeline:
    """Orchestrates the execution of analysis modules."""

    def __init__(self, structure: StructureType = None):
        """Initialize with an optional pipeline structure."""
        self.structure = structure
        self._step_count = 0
        self._step_results = {}
        self._step_timing = {}

    def run(self, input_data) -> PipelineResult:
        """Execute the pipeline structure on input data."""
        import time

        if self.structure is None:
            raise ValueError("Pipeline structure not defined")

        self._step_count = 0
        self._step_results = {}
        self._step_timing = {}

        start_time = time.time()
        result = self._process_structure(self.structure, input_data)
        total_time = time.time() - start_time

        logger.info(f"Pipeline completed in {total_time:.2f} seconds")

        return PipelineResult(
            final_output=result,
            step_outputs=self._step_results,
            execution_time=self._step_timing
        )

    def _process_structure(self, structure, data, step_path=""):
        """Process any structure recursively."""
        import time

        # Case: single module (leaf node)
        if isinstance(structure, AnalysisModule):
            step_name = structure.name
            if step_path:
                step_name = f"{step_path}.{step_name}"

            logger.info(f"Executing module: {step_name}")
            start_time = time.time()
            result = structure.run(data)
            elapsed = time.time() - start_time

            self._step_count += 1
            self._step_results[step_name] = result
            self._step_timing[step_name] = elapsed

            logger.info(f"Module {step_name} completed in {elapsed:.2f} seconds")
            return result

        # Case: dictionary with named module
        elif isinstance(structure, dict) and "module" in structure:
            module = structure["module"]
            step_name = structure.get("name", module.name)
            if step_path:
                step_name = f"{step_path}.{step_name}"

            logger.info(f"Executing named module: {step_name}")
            start_time = time.time()
            result = module.run(data)
            elapsed = time.time() - start_time

            self._step_count += 1
            self._step_results[step_name] = result
            self._step_timing[step_name] = elapsed

            logger.info(f"Module {step_name} completed in {elapsed:.2f} seconds")
            return result

        # Case: sequential flow [A, B, ...]
        elif isinstance(structure, list):
            current_data = data
            next_path = f"{step_path}.seq" if step_path else "seq"

            for i, step in enumerate(structure):
                step_id = f"{next_path}[{i}]"
                current_data = self._process_structure(step, current_data, step_id)

            return current_data

        # Case: parallel branches (A, B, ...)
        elif isinstance(structure, tuple):
            results = []
            next_path = f"{step_path}.par" if step_path else "par"

            for i, branch in enumerate(structure):
                step_id = f"{next_path}[{i}]"
                results.append(self._process_structure(branch, data, step_id))

            return tuple(results)

        else:
            raise TypeError(f"Unsupported structure type: {type(structure)}")

    def __str__(self):
        if self.structure:
            return f"AnalysisPipeline(steps={self._count_steps(self.structure)})"
        return "AnalysisPipeline(empty)"

    def _count_steps(self, structure):
        """Count the number of analysis modules in the structure."""
        if isinstance(structure, AnalysisModule):
            return 1
        elif isinstance(structure, dict) and "module" in structure:
            return 1
        elif isinstance(structure, (list, tuple)):
            return sum(self._count_steps(item) for item in structure)
        return 0


#################
# BUILDER CLASS #
#################

class Branch:
    """Builder for creating pipeline branches."""

    def __init__(self):
        self.steps = []

    def then(self, module: Union[AnalysisModule, Dict]) -> 'Branch':
        """Add a sequential step to this branch."""
        self.steps.append(module)
        return self

    def branch(self, *branches: 'Branch') -> 'Branch':
        """Create parallel branches."""
        # Extract the steps from each branch
        branch_steps = [branch.build() for branch in branches]
        if not all(branch_steps):
            raise ValueError("Cannot add empty branches")

        self.steps.append(tuple(branch_steps))
        return self

    def build(self):
        """Build the branch structure."""
        if not self.steps:
            return None
        if len(self.steps) == 1:
            return self.steps[0]
        return self.steps


class PipelineBuilder:
    """Builder for creating analysis pipelines."""

    def __init__(self):
        self.branch = Branch()

    def then(self, module: Union[AnalysisModule, Dict]) -> 'PipelineBuilder':
        """Add a sequential step to the pipeline."""
        self.branch.then(module)
        return self

    def branch(self, *branches: Branch) -> 'PipelineBuilder':
        """Create parallel branches in the pipeline."""
        self.branch.branch(*branches)
        return self

    def build(self) -> AnalysisPipeline:
        """Build the pipeline from the defined structure."""
        structure = self.branch.build()
        if not structure:
            raise ValueError("Cannot build empty pipeline")
        return AnalysisPipeline(structure)


def create_pipeline() -> PipelineBuilder:
    """Factory function to create a new pipeline builder."""
    return PipelineBuilder()


def create_branch() -> Branch:
    """Factory function to create a new branch."""
    return Branch()


#############################
# EXECUTION HELPER FUNCTIONS #
#############################

def execute_parallel(modules: List[AnalysisModule], data, max_workers: int = None):
    """Execute multiple modules in parallel and return results."""
    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        future_to_module = {
            executor.submit(module.run, data): module
            for module in modules
        }

        results = {}
        for future in future_to_module:
            module = future_to_module[future]
            try:
                result = future.result()
                results[module.name] = result
            except Exception as exc:
                logger.error(f'{module.name} generated an exception: {exc}')
                results[module.name] = None

    return results


class AnalysisModuleFactory:
    """Factory for creating common analysis module combinations."""

    @staticmethod
    def create(
            computation: ComputationModule,
            input_handler: InputHandler = None,
            output_handler: OutputHandler = None,
            name: str = None
    ) -> AnalysisModule:
        """Create an analysis module with the given components."""
        input_handler = input_handler or PassthroughHandler()
        output_handler = output_handler or DataFrameOutputHandler()
        return AnalysisModule(
            input_handler=input_handler,
            computation=computation,
            output_handler=output_handler,
            name=name or computation.__class__.__name__
        )


from typing import Tuple

from clat.util import time_util

from clat.util.connection import Connection


class EyeCalibrationParameters:
    def __init__(self, parameters=None):
        if parameters is None:
            self.parameters = [
                ('xper_right_iscan_mapping_algorithm_parameter', 3, None),
                ('xper_right_iscan_mapping_algorithm_parameter', 2, None),
                ('xper_right_iscan_mapping_algorithm_parameter', 1, None),
                ('xper_right_iscan_mapping_algorithm_parameter', 0, None),
                ('xper_left_iscan_mapping_algorithm_parameter', 0, None),
                ('xper_left_iscan_mapping_algorithm_parameter', 1, None),
                ('xper_left_iscan_mapping_algorithm_parameter', 2, None),
                ('xper_left_iscan_mapping_algorithm_parameter', 3, None),
                ('xper_right_iscan_eye_zero', 0, None),
                ('xper_right_iscan_eye_zero', 1, None),
                ('xper_left_iscan_eye_zero', 0, None),
                ('xper_left_iscan_eye_zero', 1, None)
            ]
        else:
            self.parameters = parameters

    @classmethod
    def read_params(cls, conn: Connection) -> 'EyeCalibrationParameters':
        temp_parameters = cls().parameters
        for i, (name, arr_ind, _) in enumerate(temp_parameters):
            query = "SELECT val FROM SystemVar WHERE name = %s AND arr_ind = %s order by tstamp desc limit 1"
            conn.execute(query, (name, arr_ind))
            result = conn.fetch_all()
            if result:
                temp_parameters[i] = (name, arr_ind, result[0][0])
        return cls(temp_parameters)

    @classmethod
    def clear_and_reset_eyecal(cls, conn: Connection) -> None:
        EyeCalibrationParameters._clear_eyecal(conn)

        # Insert new data
        insert_query = """
            INSERT INTO SystemVar (name, arr_ind, tstamp, val) VALUES
            ('xper_right_iscan_mapping_algorithm_parameter', 3, 1097000000000340, '1'),
            ('xper_right_iscan_mapping_algorithm_parameter', 2, 1097000000000340, '0'),
            ('xper_right_iscan_mapping_algorithm_parameter', 1, 1097000000000340, '0'),
            ('xper_right_iscan_mapping_algorithm_parameter', 0, 1097000000000340, '1'),
            ('xper_left_iscan_mapping_algorithm_parameter', 0, 1097000000000340, '1'),
            ('xper_left_iscan_mapping_algorithm_parameter', 1, 1097000000000340, '0'),
            ('xper_left_iscan_mapping_algorithm_parameter', 2, 1097000000000340, '0'),
            ('xper_left_iscan_mapping_algorithm_parameter', 3, 1097000000000340, '1'),
            ('xper_right_iscan_eye_zero', 0, 1097000000000320, '0'),
            ('xper_right_iscan_eye_zero', 1, 1097000000000320, '0'),
            ('xper_left_iscan_eye_zero', 0, 1097000000000320, '0'),
            ('xper_left_iscan_eye_zero', 1, 1097000000000320, '0')
        """
        conn.execute(insert_query)

    @classmethod
    def _clear_eyecal(cls, conn: Connection):
        # Clear existing data
        delete_queries = [
            "DELETE FROM SystemVar WHERE name = 'xper_left_iscan_eye_zero'",
            "DELETE FROM SystemVar WHERE name = 'xper_right_iscan_eye_zero'",
            "DELETE FROM SystemVar WHERE name = 'xper_left_iscan_mapping_algorithm_parameter'",
            "DELETE FROM SystemVar WHERE name = 'xper_right_iscan_mapping_algorithm_parameter'"
        ]
        for query in delete_queries:
            conn.execute(query)

    def write_params(self, conn: Connection, tstamp: int = time_util.now()) -> None:
        EyeCalibrationParameters._clear_eyecal(conn)

        query = "INSERT INTO SystemVar (name, arr_ind, tstamp, val) VALUES (%s, %s, %s, %s)"
        for name, arr_ind, val in self.parameters:
            conn.execute(query, (name, arr_ind, tstamp, val))

    def volt_to_degree(self, volts_left_right: Tuple[Tuple[float], Tuple[float]]) -> Tuple[Tuple[float, float], Tuple[float,float]]:
        # Retrieve calibration parameters for left eye
        Sxh_left = self._get_param_value('xper_left_iscan_mapping_algorithm_parameter', 0)
        Sxv_left = self._get_param_value('xper_left_iscan_mapping_algorithm_parameter', 1)
        Syh_left = self._get_param_value('xper_left_iscan_mapping_algorithm_parameter', 2)
        Syv_left = self._get_param_value('xper_left_iscan_mapping_algorithm_parameter', 3)

        # Retrieve calibration parameters for right eye
        Sxh_right = self._get_param_value('xper_right_iscan_mapping_algorithm_parameter', 0)
        Sxv_right = self._get_param_value('xper_right_iscan_mapping_algorithm_parameter', 1)
        Syh_right = self._get_param_value('xper_right_iscan_mapping_algorithm_parameter', 2)
        Syv_right = self._get_param_value('xper_right_iscan_mapping_algorithm_parameter', 3)

        # Retrieve eyeZero values for left and right eyes
        eyeZero_left_x = self._get_param_value('xper_left_iscan_eye_zero', 0)
        eyeZero_left_y = self._get_param_value('xper_left_iscan_eye_zero', 1)
        eyeZero_right_x = self._get_param_value('xper_right_iscan_eye_zero', 0)
        eyeZero_right_y = self._get_param_value('xper_right_iscan_eye_zero', 1)

        # Calculate degree for left eye
        left_volt_x, left_volt_y = volts_left_right[0]
        left_degree_y = ((left_volt_y - eyeZero_left_y) * Sxh_left - (left_volt_x - eyeZero_left_x) * Sxv_left) / (
                Syv_left * Sxh_left - Syh_left * Sxv_left)
        left_degree_x = ((left_volt_x - eyeZero_left_x) - left_degree_y * Syh_left) / Sxh_left

        # Calculate degree for right eye
        right_volt_x, right_volt_y = volts_left_right[1]
        right_degree_y = ((right_volt_y - eyeZero_right_y) * Sxh_right - (
                right_volt_x - eyeZero_right_x) * Sxv_right) / (Syv_right * Sxh_right - Syh_right * Sxv_right)
        right_degree_x = ((right_volt_x - eyeZero_right_x) - right_degree_y * Syh_right) / Sxh_right

        # Clamp values to a maximum of 90
        left_degree_x = min(left_degree_x, 90)
        left_degree_y = min(left_degree_y, 90)
        right_degree_x = min(right_degree_x, 90)
        right_degree_y = min(right_degree_y, 90)

        return (left_degree_x, left_degree_y), (right_degree_x, right_degree_y)

    def _get_param_value(self, name, arr_ind):
        for param_name, param_index, param_value in self.parameters:
            if param_name == name and param_index == arr_ind:
                return float(param_value)
        return None  # Or handle the case where the parameter is not found

    def serialize(self) -> str:
        """Serializes the parameters to a text string with each parameter on a new line."""
        serialized_data = '\n'.join([f"{name},{arr_ind},{val}" for name, arr_ind, val in self.parameters])
        return serialized_data

    @classmethod
    def deserialize(cls, serialized_data: str) -> 'EyeCalibrationParameters':
        """Deserializes the text string back into an EyeCalibrationParameters object."""
        parameters = []
        for param_str in serialized_data.strip().split('\n'):
            name, arr_ind, val = param_str.split(',')
            parameters.append((name, int(arr_ind), val))
        return cls(parameters)

    def __str__(self):
        return '\n'.join(f'Name: {name}, Arr_ind: {arr_ind}, Value: {val}' for name, arr_ind, val in self.parameters)



import numpy as np
import xmltodict
from pandas import DataFrame
from clat.util.time_util import When
from dataclasses import dataclass


def beh_msgs_during_trial(beh_msg: DataFrame, when: When):
    later_than_start = beh_msg['tstamp'] >= when.start
    before_stop = beh_msg['tstamp'] <= when.stop
    return np.logical_and(later_than_start, before_stop)


def contains_choice_success(beh_msg: DataFrame, when: When) -> bool:
    # beh_msgs = beh_msg['type'].where(beh_msgs_during_trial(beh_msg, when))
    beh_msgs = beh_msg['type'][beh_msgs_during_trial(beh_msg, when)]
    if beh_msgs.isin(["ChoiceSelectionSuccess"]).any():
        return True
    else:
        return False


def contains_calibration(beh_msg: DataFrame, when: When) -> bool:
    msg_type = beh_msg['type'].where(beh_msgs_during_trial(beh_msg, when))
    if msg_type.isin(["CalibrationPointSetup"]).any():
        return True
    else:
        return False


def get_stim_spec_id(beh_msg: DataFrame, when: When):
    def _get_trial_message(beh_msg, when: When):
        msgs_during_trial = beh_msg["msg"].where(beh_msgs_during_trial(beh_msg, when))
        trial_msg = [msg for msg in msgs_during_trial if "TrialMessage" in str(msg)][0]
        return trial_msg

    def _parse_stim_spec_from_trial_message(trial_msg):
        return int(xmltodict.parse(trial_msg)['TrialMessage']['stimSpecId'])

    trial_msg = _get_trial_message(beh_msg, when)
    stim_spec_id = _parse_stim_spec_from_trial_message(trial_msg)
    return stim_spec_id


def get_stim_spec_data(beh_msg: DataFrame, stim_spec, when: When):
    stim_spec_id = get_stim_spec_id(beh_msg, when)
    stim_spec_data_xml = stim_spec[stim_spec['id'] == stim_spec_id]['util']
    return stim_spec_data_xml.item()


@dataclass
class Coordinates2D:
    x: float
    y: float

    def __init__(self, **entries):
        """Allows initiation of Coordinates2D object with a dict with x and y component"""
        self.__dict__.update(entries)


@dataclass
class EyeLocation:
    left: Coordinates2D
    right: Coordinates2D


def get_eye_location_volts(beh_msg_eye: DataFrame, when: When) -> [EyeLocation]:
    """Returns a list of EyeLocation objects given the beh_msg_eye table and timestamps to search within"""
    eye_msgs_during_trial = beh_msg_eye['msg'].where(beh_msgs_during_trial(beh_msg_eye, when))
    left_volts = [xmltodict.parse(msg)['EyeDeviceMessage']['volt'] for msg in eye_msgs_during_trial if
                  "leftIscan" in str(msg)]
    right_volts = [xmltodict.parse(msg)['EyeDeviceMessage']['volt'] for msg in eye_msgs_during_trial if
                   "rightIscan" in str(msg)]
    locations = _dicts_to_eye_location(left_volts, right_volts)
    return locations


def _dicts_to_eye_location(left_xy_dict, right_xy_dict):
    """Converts two dictionaries composing of x:float and y:float into EyeLocation composed of Coordinates2D"""
    locations = [EyeLocation(Coordinates2D(**left_volt), Coordinates2D(**right_volt)) for left_volt, right_volt in
                 zip(left_xy_dict, right_xy_dict)]
    return locations



import os
import shutil
import tkinter as tk
from tkinter import simpledialog

import tkfilebrowser


class IntanFileStitcher:
    def __init__(self, folder_paths):
        self.folder_paths = sorted(folder_paths)

    def read_append_write(self, filename, output_folder):
        with open(os.path.join(output_folder, filename), 'wb') as output_file:
            for folder in self.folder_paths:
                input_file_path = os.path.join(folder, filename)
                with open(input_file_path, 'rb') as input_file:
                    shutil.copyfileobj(input_file, output_file)

    def append_notes(self, filename, output_folder):
        cumulative_last_index = 0
        with open(os.path.join(output_folder, filename), 'w') as output_file:
            for folder in self.folder_paths:
                input_file_path = os.path.join(folder, filename)
                local_last_index = 0  # Last index within the current file
                with open(input_file_path, 'r') as input_file:
                    lines = input_file.readlines()
                    for line in lines:
                        line = line.strip()
                        if not line:  # Skip empty lines
                            continue
                        index, timestamp, info = line.split(", ")
                        new_index = int(index) + cumulative_last_index
                        output_file.write(f"{new_index}, {timestamp}, {info}\n\n")  # Added two extra newlines
                        local_last_index = new_index  # Update the last index for the current file
                cumulative_last_index = local_last_index  # Update the cumulative last index for the next file

    def copy_auxiliary_files(self, filename, output_folder):
        source_path = os.path.join(self.folder_paths[0], filename)
        destination_path = os.path.join(output_folder, filename)
        shutil.copyfile(source_path, destination_path)

    def stitch_files(self, output_folder):
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)

        files_to_stitch = ['amplifier.dat', 'digitalin.dat']
        for filename in files_to_stitch:
            self.read_append_write(filename, output_folder)

        self.append_notes('notes.txt', output_folder)

        auxiliary_files = ['info.rhd', 'settings.xml']
        for filename in auxiliary_files:
            self.copy_auxiliary_files(filename, output_folder)

        self.create_merge_info(output_folder)

    def create_merge_info(self, output_folder):
        with open(os.path.join(output_folder, "mergeinfo.txt"), 'w') as f:
            for folder in self.folder_paths:
                folder_name = os.path.basename(folder)
                f.write(f"{folder_name}\n")

    def stitch_spike_dat(self, output_folder):
        """
        Stitch multiple spike.dat files together, updating timestamps correctly.
        This handles the special binary format of spike.dat files.
        """
        # Initialize structures to store data from all folders
        all_spikes = {}
        last_sample_rate = None

        # First read all spike data from all folders
        file_offsets = {}  # To track time offsets for each file
        max_timestamp = 0

        print("Processing spike.dat files...")

        # First pass: read all files and determine timestamp offsets
        for i, folder in enumerate(self.folder_paths):
            input_file_path = os.path.join(folder, 'spike.dat')
            if not os.path.exists(input_file_path):
                print(f"Warning: spike.dat not found in {folder}")
                continue

            if i == 0:
                file_offsets[folder] = 0  # First file has no offset
            else:
                file_offsets[folder] = max_timestamp  # Offset for subsequent files

            # Read the spike file and extract data
            spikes, sample_rate = self.read_intan_spike_file(input_file_path, no_artifacts=False)

            if last_sample_rate is not None and sample_rate != last_sample_rate:
                print(f"Warning: Sample rate mismatch between folders: {last_sample_rate} vs {sample_rate}")

            last_sample_rate = sample_rate

            # Process each channel in the spike data
            for channel_data in spikes:
                channel_name = channel_data[0]  # Native channel name
                custom_name = channel_data[1]  # Custom channel name
                timestamps = channel_data[2]  # Timestamps
                spike_ids = channel_data[3]  # Spike IDs

                # Get snapshots if they exist
                snapshots = channel_data[4] if len(channel_data) > 4 else None

                # Update timestamps based on file offset
                adjusted_timestamps = [t + file_offsets[folder] / sample_rate for t in timestamps]

                # Find the maximum timestamp for the next offset calculation
                if adjusted_timestamps:
                    current_max = max(adjusted_timestamps)
                    max_timestamp = max(max_timestamp, current_max * sample_rate)

                # Initialize channel in all_spikes if not already present
                if channel_name not in all_spikes:
                    all_spikes[channel_name] = {
                        'custom_name': custom_name,
                        'timestamps': [],
                        'spike_ids': [],
                        'snapshots': [] if snapshots else None
                    }

                # Append data for this channel
                all_spikes[channel_name]['timestamps'].extend(adjusted_timestamps)
                all_spikes[channel_name]['spike_ids'].extend(spike_ids)
                if snapshots:
                    all_spikes[channel_name]['snapshots'].extend(snapshots)

        # Now write the combined data to a new spike.dat file
        output_file_path = os.path.join(output_folder, 'spike.dat')

        if all_spikes and last_sample_rate:
            self.write_intan_spike_file(output_file_path, all_spikes, last_sample_rate)
            print(f"Created stitched spike.dat file at {output_file_path}")
        else:
            print("No spike data found to stitch.")
def open_gui():
    root = tk.Tk()
    root.withdraw()  # Hide the main window

    default_folder = "/home/r2_allen/Documents/JulieIntanData/Cortana"
    # default_folder = "/run/user/1003/gvfs/sftp:host=172.30.6.58/home/connorlab/Documents/IntanData"

    folder_paths = tkfilebrowser.askopendirnames(initialdir=default_folder, title="Select Folders to Stitch")
    folder_paths = list(folder_paths)  # Convert tuple to list

    if folder_paths:  # If folders were selected
        output_folder_name = simpledialog.askstring("Output Folder", "Enter the name for the output folder:")

        if output_folder_name:  # If an output folder name was provided
            # Get the parent directory of each selected folder
            parent_directories = [os.path.dirname(folder) for folder in folder_paths]

            # Get the common parent directory of all selected folders
            common_parent_directory = os.path.commonprefix(parent_directories)

            # Append the output folder name to the common parent directory
            final_output_folder_path = os.path.join(common_parent_directory, output_folder_name)

            # Run the stitcher
            stitcher = IntanFileStitcher(folder_paths)
            stitcher.stitch_files(final_output_folder_path)
def main():
    open_gui()

if __name__ == '__main__':
    main()


#! /bin/env python
#
# Michael Gibson 23 April 2015
# Modified Adrian Foy Sep 2018


def get_bytes_per_data_block(header):
    """Calculates the number of bytes in each 60 or 128 sample datablock."""

    # Each data block contains 60 or 128 amplifier samples.
    bytes_per_block = header['num_samples_per_data_block'] * 4  # timestamp data
    bytes_per_block = bytes_per_block + header['num_samples_per_data_block'] * 2 * header['num_amplifier_channels']

    # Auxiliary inputs are sampled 4x slower than amplifiers
    bytes_per_block = bytes_per_block + (header['num_samples_per_data_block'] / 4) * 2 * header['num_aux_input_channels']

    # Supply voltage is sampled 60 or 128x slower than amplifiers
    bytes_per_block = bytes_per_block + 1 * 2 * header['num_supply_voltage_channels']

    # Board analog inputs are sampled at same rate as amplifiers
    bytes_per_block = bytes_per_block + header['num_samples_per_data_block'] * 2 * header['num_board_adc_channels']

    # Board digital inputs are sampled at same rate as amplifiers
    if header['num_board_dig_in_channels'] > 0:
        bytes_per_block = bytes_per_block + header['num_samples_per_data_block'] * 2

    # Board digital outputs are sampled at same rate as amplifiers
    if header['num_board_dig_out_channels'] > 0:
        bytes_per_block = bytes_per_block + header['num_samples_per_data_block'] * 2

    # Temp sensor is sampled 60 or 128x slower than amplifiers
    if header['num_temp_sensor_channels'] > 0:
        bytes_per_block = bytes_per_block + 1 * 2 * header['num_temp_sensor_channels']

    return bytes_per_block
  


import os
from typing import Tuple, List, Dict


def map_task_id_to_epochs_with_livenotes(livenotes_data: str,
                                         marker_channel_time_indices: List[Tuple[int, int]],
                                         require_trial_complete=True,
                                         is_output_first_instance=False) -> Dict[
    int, Tuple[int, int]]:
    """
    If require_trial_complete is True:
    Map unique task IDs to epochs with a following 'Trial Complete' event,
    such that if there are repetitions of a task id in the livenotes, only the complete instance will
    be returned.

    If require_trial_complete is False:
    Map task IDs to epochs without requiring a following 'Trial Complete' event. If a taskId
    has multiple instances in the livenotes, output_first_instance will determine whether the first
    instance or the last instance will be returned. If output_first_instance is True, the first instance
    will be returned, otherwise the last instance will be returned.
    """
    data = read_livenotes(livenotes_data)
    events = parse_livenotes_to_events(data)
    task_ids = filter_for_task_ids(events)
    task_ids.sort()

    result = {}

    # Loop through each task_id and find the closest marker channel to it
    for tstamp, task_id in task_ids:
        closest_start = None
        closest_end = None
        following_event = None

        idx = events.index((tstamp, str(task_id)))  # Find the index of this task_id in the original events list
        if idx < len(events) - 1:
            following_event = events[idx + 1][1]  # Get the following event


        if require_trial_complete:
            # Only proceed if the following event is 'Trial Complete'
            if following_event == 'Trial Complete':
                for epoch_start, epoch_end in marker_channel_time_indices:
                    if closest_start is None or is_epoch_closer(closest_start, epoch_start, tstamp):
                        closest_start = epoch_start
                        closest_end = epoch_end

                if closest_start is not None and result.get(task_id) is None:
                    result[task_id] = (closest_start, closest_end)
        else:
            for epoch_start, epoch_end in marker_channel_time_indices:
                if closest_start is None or is_epoch_closer(closest_start, epoch_start, tstamp):
                    closest_start = epoch_start
                    closest_end = epoch_end

            if is_output_first_instance:
                if closest_start is not None and result.get(task_id) is None:
                    result[task_id] = (closest_start, closest_end)
            else:
                if closest_start is not None:
                    # will override the previous instance of the task_id
                    # so that the last instance will be returned at the end
                    result[task_id] = (closest_start, closest_end)

    return result


def is_epoch_closer(closest_start, epoch_start, tstamp):
    return abs(tstamp - epoch_start) < abs(tstamp - closest_start)


def is_no_closest_start_yet(closest_start):
    return closest_start is None


def map_unique_task_id_to_epochs_with_livenotes(livenotes_data: str,
                                                marker_channel_time_indices: List[tuple]) -> dict[
    int, Tuple[int, int]]:
    """
    This functions requires task_ids to be unique in the livenotes file. If there are multiple task_ids in the livenotes, it will output the
    all instances of the task_id: (start, end) in the marker_channel_time_indices

    Params:
    livenotes_data: live_notes file in the form a path or the file string itself
    marker_channel_time_indices: list of tuples (start, end) where start and end are the start and end time indices of the stimulus
    based on marker_channel data

    Returns:
    mapping of the stim_ids in the livenotes with the real marker-channel based tuples (start, end)
    based on closest matching between timestamp in livenotes and start time in marker-channel data

    """
    data = read_livenotes(livenotes_data)

    tstamp_and_events_from_livenotes = parse_livenotes_to_events(data)
    tstamp_and_stim_id_from_livenotes = filter_for_task_ids(tstamp_and_events_from_livenotes)

    # Sort the tstamp_and_stim_id_from_livenotes by tstamp
    tstamp_and_stim_id_from_livenotes.sort()

    # Initialize the dictionary to store the result
    result = {}

    # For each tuple in time_indices, find the one with the closest tstamp
    for start, end in marker_channel_time_indices:
        # Find the record with the tstamp closest to start
        closest_tstamp = None
        closest_stim_id = None
        for tstamp, stim_id in tstamp_and_stim_id_from_livenotes:
            if stim_id not in result and (closest_tstamp is None or abs(tstamp - start) < abs(closest_tstamp - start)):
                closest_tstamp = tstamp
                closest_stim_id = stim_id

        # If no match is found, raise an error
        if closest_stim_id is None:
            print(f"No match found for start time {start} found in marker channels")

        # Otherwise, add it to the result
        result[closest_stim_id] = (start, end)

    return result


def filter_for_task_ids(tstamp_and_events_from_livenotes: List[Tuple[float, str]]) -> List[Tuple[float, int]]:
    filtered = []
    for tstamp, event in tstamp_and_events_from_livenotes:
        try:
            stim_id = int(event)
            filtered.append((tstamp, stim_id))
        except ValueError:
            continue
    return filtered


def parse_livenotes_to_events(data: str) -> List[Tuple[int, str]]:
    # Convert the raw text data into a list of tuples (tstamp, stim_id)
    tstamp_and_events_from_livenotes = []
    for line in data.strip().split('\n\n'):
        try:
            parts = line.split(',')
            tstamp = int(parts[0].strip())
            event = parts[2].strip()
            tstamp_and_events_from_livenotes.append((tstamp, event))
        except IndexError:
            print(f"Error parsing line {line}")
            continue

    return tstamp_and_events_from_livenotes


def read_livenotes(livenotes_data: str) -> str:
    # Check if the input is a file path
    if os.path.isfile(livenotes_data):
        with open(livenotes_data, 'r') as file:
            data = file.read()
    else:
        data = livenotes_data
    return data



from __future__ import annotations

from collections import OrderedDict

import pandas as pd

from clat.util.connection import Connection
from clat.util.time_util import When

"""
Handles compilation of data while using the timestamps of trialStart and trialStop
as the primary keys for the data.
"""
class Field:
    def __init__(self, name: str = None):
        if name is None:
            self.name = type(self).__name__
        else:
            self.name = name

    def get(self, when: When):
        raise NotImplementedError("Not Implemented")


class FieldList(list[Field]):
    """List of Field types"""

    def get_df(self):
        df = pd.DataFrame(columns=self.get_names())
        return df

    def get_names(self):
        return [field.name for field in self]

    def get_data(self, trial_tstamps: list[When]) -> pd.DataFrame:
        return self._get_data_from_trials(self, trial_tstamps)

    def _get_data_from_trials(self, fields: FieldList, trial_tstamps: list[When]) -> pd.DataFrame:
        data = []
        for i, when in enumerate(trial_tstamps):
            if i % 100 == 0:
                print("working on " + str(i) + " out of " + str(len(trial_tstamps)))
            field_values = [field.get(when) for field in fields]
            names = fields.get_names()
            new_row = OrderedDict(zip(names, field_values))
            data.append(new_row)

        return pd.DataFrame(data)


def get_data_from_trials(fields: FieldList, trial_tstamps: list[When]) -> pd.DataFrame:
    data = []
    for i, when in enumerate(trial_tstamps):
        if i % 100 == 0:
            print("working on " + str(i) + " out of " + str(len(trial_tstamps)))
        field_values = [field.get(when) for field in fields]
        names = fields.get_names()
        new_row = OrderedDict(zip(names, field_values))
        data.append(new_row)

    return pd.DataFrame(data)

class DatabaseField(Field):
    def __init__(self, conn: Connection, name: str = None):
        super().__init__(name)
        self.conn = conn

    def get(self, when: When):
        raise NotImplementedError("Not Implemented")






from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='ConnorLabAnalysisTools',
    version='1.0.0',
    description='Tools for analyzing data in the Connor Lab',
    # long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Allen Chen',
    author_email='allenmuhanchen@gmail.com',
    url='https://github.com/EdConnorLab/ConnorLabAnalysisTools',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=required,
)




import xmltodict

from clat.util.connection import Connection


class TaskIdCollector:
    def __init__(self, conn: Connection):
        self.conn = conn

    def collect_task_ids(self):
        print("Collecting task IDs")
        self.conn.execute(
            "SELECT msg FROM BehMsg WHERE type = 'SlideOff'"
        )
        slide_off_msgs = self.conn.fetch_all()
        task_ids = []
        for msg in slide_off_msgs:
            slide_event_dict = xmltodict.parse(msg[0])
            task_id = int(slide_event_dict['SlideEvent']['taskId'])
            task_ids.append(task_id)
        return task_ids


class PngSlideIdCollector:
    """
    For Julie's experiment
    """
    def __init__(self, conn: Connection):
        self.conn = conn

    def collect_complete_task_ids(self, time_range: tuple):
        print("Collecting task IDs from Complete Trials (SlideOff Events)")

        # Unpack the start and end times from the tuple
        start_time, end_time = time_range

        # Fetch all messages in the time range
        self.conn.execute(
            f"SELECT tstamp, type, msg FROM BehMsg WHERE tstamp BETWEEN {start_time} AND {end_time} ORDER BY tstamp ASC"
        )

        all_msgs = self.conn.fetch_all()

        slide_off = None
        task_ids = []
        for tstamp, msg_type, msg in all_msgs:
            if msg_type == 'SlideOff':
                slide_off = xmltodict.parse(msg)
            elif msg_type == 'TrialComplete' and slide_off:
                try:
                    task_id = int(slide_off['PngSlideEvent']['taskId'])
                    task_ids.append(task_id)
                except KeyError:
                    print("Ignored a non-PNG slide event")
                slide_off = None
            elif msg_type == 'TrialStop':
                slide_off = None

        return task_ids





#! /bin/env python
#
# Michael Gibson 17 July 2015
# Modified Adrian Foy January 2023

import os
import sys
import time

import numpy as np

from clat.intan.rhd.intanutil.data_to_result import data_to_result
from clat.intan.rhd.intanutil.get_bytes_per_data_block import get_bytes_per_data_block
from clat.intan.rhd.intanutil.notch_filter import notch_filter
from clat.intan.rhd.intanutil.read_header import read_header
from clat.intan.rhd.intanutil.read_one_data_block import read_one_data_block


def read_data(filename):
    """Reads Intan Technologies RHD2000 data file generated by evaluation board GUI.
    
    Data are returned in a dictionary, for future extensibility.
    """

    tic = time.time()
    fid = open(filename, 'rb')
    filesize = os.path.getsize(filename)

    header = read_header(fid)

    print('Found {} amplifier channel{}.'.format(header['num_amplifier_channels'], plural(header['num_amplifier_channels'])))
    print('Found {} auxiliary input channel{}.'.format(header['num_aux_input_channels'], plural(header['num_aux_input_channels'])))
    print('Found {} supply voltage channel{}.'.format(header['num_supply_voltage_channels'], plural(header['num_supply_voltage_channels'])))
    print('Found {} board ADC channel{}.'.format(header['num_board_adc_channels'], plural(header['num_board_adc_channels'])))
    print('Found {} board digital input channel{}.'.format(header['num_board_dig_in_channels'], plural(header['num_board_dig_in_channels'])))
    print('Found {} board digital output channel{}.'.format(header['num_board_dig_out_channels'], plural(header['num_board_dig_out_channels'])))
    print('Found {} temperature sensors channel{}.'.format(header['num_temp_sensor_channels'], plural(header['num_temp_sensor_channels'])))
    print('')

    # Determine how many samples the data file contains.
    bytes_per_block = get_bytes_per_data_block(header)

    # How many data blocks remain in this file?
    data_present = False
    bytes_remaining = filesize - fid.tell()
    if bytes_remaining > 0:
        data_present = True

    if bytes_remaining % bytes_per_block != 0:
        raise Exception('Something is wrong with file size : should have a whole number of data blocks')

    num_data_blocks = int(bytes_remaining / bytes_per_block)

    num_amplifier_samples = header['num_samples_per_data_block'] * num_data_blocks
    num_aux_input_samples = int((header['num_samples_per_data_block'] / 4) * num_data_blocks)
    num_supply_voltage_samples = 1 * num_data_blocks
    num_board_adc_samples = header['num_samples_per_data_block'] * num_data_blocks
    num_board_dig_in_samples = header['num_samples_per_data_block'] * num_data_blocks
    num_board_dig_out_samples = header['num_samples_per_data_block'] * num_data_blocks

    record_time = num_amplifier_samples / header['sample_rate']

    if data_present:
        print('File contains {:0.3f} seconds of data.  Amplifiers were sampled at {:0.2f} kS/s.'.format(record_time, header['sample_rate'] / 1000))
    else:
        print('Header file contains no data.  Amplifiers were sampled at {:0.2f} kS/s.'.format(header['sample_rate'] / 1000))

    if data_present:
        # Pre-allocate memory for data.
        print('')
        print('Allocating memory for data...')

        data = {}
        if (header['version']['major'] == 1 and header['version']['minor'] >= 2) or (header['version']['major'] > 1):
            data['t_amplifier'] = np.zeros(num_amplifier_samples, dtype=np.int_)
        else:
            data['t_amplifier'] = np.zeros(num_amplifier_samples, dtype=np.uint)

        data['amplifier_data'] = np.zeros([header['num_amplifier_channels'], num_amplifier_samples], dtype=np.uint)
        data['aux_input_data'] = np.zeros([header['num_aux_input_channels'], num_aux_input_samples], dtype=np.uint)
        data['supply_voltage_data'] = np.zeros([header['num_supply_voltage_channels'], num_supply_voltage_samples], dtype=np.uint)
        data['temp_sensor_data'] = np.zeros([header['num_temp_sensor_channels'], num_supply_voltage_samples], dtype=np.uint)
        data['board_adc_data'] = np.zeros([header['num_board_adc_channels'], num_board_adc_samples], dtype=np.uint)
        
        # by default, this script interprets digital events (digital inputs and outputs) as booleans
        # if unsigned int values are preferred(0 for False, 1 for True), replace the 'dtype=np.bool_' argument with 'dtype=np.uint' as shown
        # the commented line below illustrates this for digital input data; the same can be done for digital out
        
        #data['board_dig_in_data'] = np.zeros([header['num_board_dig_in_channels'], num_board_dig_in_samples], dtype=np.uint)
        data['board_dig_in_data'] = np.zeros([header['num_board_dig_in_channels'], num_board_dig_in_samples], dtype=np.bool_)
        data['board_dig_in_raw'] = np.zeros(num_board_dig_in_samples, dtype=np.uint)
        
        data['board_dig_out_data'] = np.zeros([header['num_board_dig_out_channels'], num_board_dig_out_samples], dtype=np.bool_)
        data['board_dig_out_raw'] = np.zeros(num_board_dig_out_samples, dtype=np.uint)

        # Read sampled data from file.
        print('Reading data from file...')

        # Initialize indices used in looping
        indices = {}
        indices['amplifier'] = 0
        indices['aux_input'] = 0
        indices['supply_voltage'] = 0
        indices['board_adc'] = 0
        indices['board_dig_in'] = 0
        indices['board_dig_out'] = 0

        print_increment = 10
        percent_done = print_increment
        for i in range(num_data_blocks):
            read_one_data_block(data, header, indices, fid)

            # Increment indices
            indices['amplifier'] += header['num_samples_per_data_block']
            indices['aux_input'] += int(header['num_samples_per_data_block'] / 4)
            indices['supply_voltage'] += 1
            indices['board_adc'] += header['num_samples_per_data_block']
            indices['board_dig_in'] += header['num_samples_per_data_block']
            indices['board_dig_out'] += header['num_samples_per_data_block']            

            fraction_done = 100 * (1.0 * i / num_data_blocks)
            if fraction_done >= percent_done:
                print('{}% done...'.format(percent_done))
                percent_done = percent_done + print_increment

        # Make sure we have read exactly the right amount of data.
        bytes_remaining = filesize - fid.tell()
        if bytes_remaining != 0: raise Exception('Error: End of file not reached.')



    # Close data file.
    fid.close()

    if (data_present):
        print('Parsing data...')

        # Extract digital input channels to separate variables.
        for i in range(header['num_board_dig_in_channels']):
            data['board_dig_in_data'][i, :] = np.not_equal(np.bitwise_and(data['board_dig_in_raw'], (1 << header['board_dig_in_channels'][i]['native_order'])), 0)

        # Extract digital output channels to separate variables.
        for i in range(header['num_board_dig_out_channels']):
            data['board_dig_out_data'][i, :] = np.not_equal(np.bitwise_and(data['board_dig_out_raw'], (1 << header['board_dig_out_channels'][i]['native_order'])), 0)

        # Scale voltage levels appropriately.
        data['amplifier_data'] = np.multiply(0.195, (data['amplifier_data'].astype(np.int32) - 32768))      # units = microvolts
        data['aux_input_data'] = np.multiply(37.4e-6, data['aux_input_data'])               # units = volts
        data['supply_voltage_data'] = np.multiply(74.8e-6, data['supply_voltage_data'])     # units = volts
        if header['eval_board_mode'] == 1:
            data['board_adc_data'] = np.multiply(152.59e-6, (data['board_adc_data'].astype(np.int32) - 32768)) # units = volts
        elif header['eval_board_mode'] == 13:
            data['board_adc_data'] = np.multiply(312.5e-6, (data['board_adc_data'].astype(np.int32) - 32768)) # units = volts
        else:
            data['board_adc_data'] = np.multiply(50.354e-6, data['board_adc_data'])           # units = volts
        data['temp_sensor_data'] = np.multiply(0.01, data['temp_sensor_data'])               # units = deg C

        # Check for gaps in timestamps.
        num_gaps = np.sum(np.not_equal(data['t_amplifier'][1:]-data['t_amplifier'][:-1], 1))
        if num_gaps == 0:
            print('No missing timestamps in data.')
        else:
            print('Warning: {0} gaps in timestamp data found.  Time scale will not be uniform!'.format(num_gaps))

        # Scale time steps (units = seconds).
        data['t_amplifier'] = data['t_amplifier'] / header['sample_rate']
        data['t_aux_input'] = data['t_amplifier'][range(0, len(data['t_amplifier']), 4)]
        data['t_supply_voltage'] = data['t_amplifier'][range(0, len(data['t_amplifier']), header['num_samples_per_data_block'])]
        data['t_board_adc'] = data['t_amplifier']
        data['t_dig'] = data['t_amplifier']
        data['t_temp_sensor'] = data['t_supply_voltage']

        # If the software notch filter was selected during the recording, apply the
        # same notch filter to amplifier data here.
        if header['notch_filter_frequency'] > 0 and header['version']['major'] < 3:
            print('Applying notch filter...')

            print_increment = 10
            percent_done = print_increment
            for i in range(header['num_amplifier_channels']):
                data['amplifier_data'][i,:] = notch_filter(data['amplifier_data'][i,:], header['sample_rate'], header['notch_filter_frequency'], 10)

                fraction_done = 100 * (i / header['num_amplifier_channels'])
                if fraction_done >= percent_done:
                    print('{}% done...'.format(percent_done))
                    percent_done += print_increment
    else:
        data = [];

    # Move variables to result struct.
    result = data_to_result(header, data, data_present)

    print('Done!  Elapsed time: {0:0.1f} seconds'.format(time.time() - tic))
    return result

def plural(n):
    """Utility function to optionally pluralize words based on the value of n.
    """

    if n == 1:
        return ''
    else:
        return 's'

if __name__ == '__main__':
    a=read_data(sys.argv[1])
    #print(a)


import datetime
import threading
from time import sleep

import mysql.connector
import pandas as pd

from clat.util import time_util
from clat.util.time_util import When, to_unix


class Connection:

    def __init__(self, database, user="xper_rw", password="up2nite", host="172.30.6.80"):
        self.database = database
        self.user = user
        self.password = password
        self.host = host
        self._connect()
        self.lock = threading.Lock()

    def _connect(self):
        self.my_cursor = None
        self.mydb = mysql.connector.connect(
            host=self.host,
            user=self.user,
            password=self.password,
            database=self.database,
            autocommit=True
        )

    def execute(self, statement, params=()):
        with self.lock:
            try:
                try:
                    self.my_cursor.fetchall()
                    self.my_cursor.close()
                except:
                    pass
                self.my_cursor = self.mydb.cursor()
            except mysql.connector.errors.OperationalError as e:
                sleep(1)
                self._connect()
                self.my_cursor = self.mydb.cursor()
            self.my_cursor.execute(statement, params)

            has_results = self.my_cursor.description is not None
            if not has_results:
                self.mydb.commit()
                self.my_cursor.close()

    def truncate(self, table_name):
        self.execute(f"TRUNCATE TABLE {table_name}")

    def fetch_one(self):
        with self.lock:
            result = None
            try:
                result = self.my_cursor.fetchone()
            except mysql.connector.errors.InternalError as e:
                if "Unread result found" in str(e):
                    self.my_cursor.fetchall()
                    result = self.my_cursor.fetchone()
            if result:
                return result[0]
            else:
                return None

    def fetch_all(self):
        with self.lock:
            result = self.my_cursor.fetchall()
            # self.my_cursor.fetchall()
            self.my_cursor.close()
            return result

    def get_beh_msg(self, when: When) -> pd.DataFrame:
        try:
            self.execute("SELECT * FROM BehMsg WHERE tstamp>= %s && tstamp<=%s", (when.start, when.stop))
            df = pd.DataFrame(self.fetch_all())
            df.columns = ['tstamp', 'type', 'msg']
        except ValueError as e:
            print("Error: BehMsg empty for time: " + str(when))
            raise ValueError
        return df

    def get_stim_spec(self, when: When) -> pd.DataFrame:
        self.execute("SELECT * FROM StimSpec WHERE id>= %s & id<=%s", (when.start, when.stop))
        df = pd.DataFrame(self.fetch_all())
        df.columns = ['id', 'spec', 'util']
        return df

    def get_stim_obj_data(self, when: When) -> pd.DataFrame:
        try:
            self.execute("SELECT * FROM StimObjData WHERE id>= %s & id<=%s", (when.start, when.stop))
            df = pd.DataFrame(self.fetch_all())
            df.columns = ['id', 'spec', 'util']
            return df
        except:
            return None

    def get_beh_msg_eye(self, when: When) -> pd.DataFrame:
        self.execute("SELECT * FROM BehMsgEye WHERE tstamp>= %s & tstamp<=%s", (when.start, when.stop))
        df = pd.DataFrame(self.fetch_all())
        df.columns = ['tstamp', 'type', 'msg']
        return df


def since_nth_most_recent_experiment(conn: Connection, n=1):
    """
    Get the start timestamp of the nth most recent experiment.

    Parameters:
    - conn: Connection object.
    - n: Specifies the nth most recent experiment. Defaults to 1 (most recent).

    Returns:
    - A When object representing the time range of the nth most recent experiment.
    """
    query = f"""
            SELECT tstamp 
            FROM BehMsg 
            WHERE type = 'ExperimentStart'
            ORDER BY tstamp DESC
            LIMIT 1 OFFSET {n - 1}
        """
    conn.execute(query)
    result = conn.fetch_all()

    if result:
        start = result[0][0]
    else:
        # Handle the case where there are less than n experiments
        raise ValueError(f"No data available for the {n}th most recent experiment.")

    # Assuming 'to_unix' is a function to convert datetime to Unix timestamp
    end = to_unix(datetime.date.fromisoformat('3022-01-01'))
    return When(start, end)


def get_time_range_for_experiment_id(conn, experiment_id):
    # Fetch all experiment IDs from the CurrentExperiments table
    query = "SELECT experiment_id FROM CurrentExperiments ORDER BY experiment_id"
    conn.execute(query)
    experiment_ids = [row[0] for row in conn.fetch_all()]

    # Check if the given experiment_id is the most recent one
    if experiment_id == experiment_ids[-1]:
        # If it's the most recent, return a When object with experiment_id and current time
        return When(experiment_id, time_util.now())
    else:
        # If it's not the most recent, find the next experiment_id chronologically
        next_experiment_id = None
        for exp_id in experiment_ids:
            if exp_id > experiment_id:
                next_experiment_id = exp_id
                break

        if next_experiment_id is not None:
            # Return a When object with the given experiment_id and the next experiment_id
            return When(experiment_id, next_experiment_id)
        else:
            raise ValueError(f"No experiment found after experiment ID {experiment_id}")



import os

import numpy as np


def epoch_using_marker_channels(digitalin_path, false_negative_correction_duration=40,
                                false_positive_correction_duration=2) -> list[tuple[int, int]]:
    """
    :param digitalin_path: path to digitalin.dat file
    :param false_negative_correction_duration: the number of samples to check before or after a state switch to identify false negatives or positives
    :return: list of tuples of start and stop indices for each epoch
    """
    digitalin = read_digitalin_file(digitalin_path)
    return get_epochs_start_and_stop_indices(digitalin[0], digitalin[1], false_negative_correction_duration,
                                             false_positive_correction_duration)


def get_epochs_start_and_stop_indices(marker1_data, marker2_data, false_negative_correction_duration=40,
                                      false_positive_correction_duration=2) -> list[tuple[int, int]]:
    epochs = []
    start_time = None
    current_marker = determine_marker_with_first_pulse(marker1_data, marker2_data)

    for i in range(max(len(marker1_data), len(marker2_data))):
        if current_marker == 1:
            # Starting an epoch for marker 1
            if marker1_data[i] and start_time is None and not false_positive(i, marker1_data,
                                                                             false_positive_correction_duration):
                start_time = i
            # Check for potential end of epoch
            elif epoch_ongoing(start_time) and is_end_of_epoch(i, marker1_data):
                # Check if it's a false negative (gap too short)
                if false_negative(i, false_negative_correction_duration, marker1_data):
                    if i % false_negative_correction_duration == 0:  # Only print occasionally
                        print(f"Detected false negative for marker 1 at time {i}")
                else:
                    # It's a genuine end of the epoch
                    epochs.append((start_time, i))
                    start_time = None
                    current_marker = 2
        else:
            # Same logic for marker 2
            if marker2_data[i] and start_time is None and not false_positive(i, marker2_data,
                                                                             false_positive_correction_duration):
                start_time = i
            elif epoch_ongoing(start_time) and is_end_of_epoch(i, marker2_data):
                if false_negative(i, false_negative_correction_duration, marker2_data):
                    if i % false_negative_correction_duration == 0:
                        print(f"Detected false negative for marker 2 at time {i}")
                else:
                    epochs.append((start_time, i))
                    start_time = None
                    current_marker = 1

    return epochs


def epoch_ongoing(start_time) -> bool:
    return start_time is not None


def is_end_of_epoch(i, marker_data) -> bool:
    try:
        return not marker_data[i + 1]
    except IndexError:
        return True


def false_positive(i, marker_data, min_duration) -> bool:
    # Check if any of the values ahead of time are False
    for j in range(i, i + min_duration):
        # If any of the values is False, that means this is a false positive
        try:
            if not marker_data[j]:
                return True
        except IndexError:
            return False


def false_negative(i, min_duration, marker_data) -> bool:
    # Check min_duration ahead of time
    for j in range(i + 1, i + min_duration + 1):
        # If any of the values is True, that means the epoch is not actually over
        try:
            if marker_data[j]:
                return True
        except IndexError:
            return False


def determine_marker_with_first_pulse(marker1_data, marker2_data) -> int:
    for i in range(len(marker1_data)):
        if marker1_data[i] and not marker2_data[i]:
            return 1
        elif marker2_data[i] and not marker1_data[i]:
            return 2
    return None


def read_digitalin_file(full_file_name) -> list[list[bool], list[bool]]:
    fid = open(full_file_name, 'rb')
    filesize = os.path.getsize(full_file_name)

    num_samples = filesize // 2  # uint16 = 2 bytes

    digital_word = np.fromfile(fid, dtype=np.uint16, count=num_samples)
    fid.close()

    # Get the digital inputs for channel 0 and 1
    digital_input_ch0 = (digital_word & (2 ** 0)) > 0
    digital_input_ch1 = (digital_word & (2 ** 1)) > 0

    return [digital_input_ch0, digital_input_ch1]


def isolate_digital_input(digital_word, ch):
    digital_input_ch = (digital_word & (2 ** ch)) > 0
    return digital_input_ch
